        TITLE   Book CpuX86.asm - 80(x)86 Cpu Definitions
        PAGE    53,160

;
; CpuX86.asm - 80(X)86 Cpu definitions
;
; This document contains file that you may use at your own risk in your own programs.
;
; This module is designed to accomodate X86 processors in the book.
;
;       Created - 27 Jun 98 - J.Leiterman
;
;       Modifications
;
;               23 Jul 2001 - JCL - Brought up to date.
;
;       Tabs = 8



        .XLIST

        .387
        .686P
        .MMX
        .model flat, C

        include KariType.inc
;        include Prolog32.inc
        include x86.inc
        include CpuX86.inc
 
        .LIST



;       Vendor Data Structure

VENDOR  STRUCT  4
        vname   BYTE    '------------'
        Id      DWORD   CPUVEN_UNKNOWN
VENDOR  ENDS



        .data
;------------------------------------------------------------------------------------------------
;       CPUID information

TheCpuInfo      CpuInfo {}

        public  SerialNum

SerialNum       dd      0,0,0

Fpu             dd      0


;       Vendor Data List

VendorList      VENDOR { "AMD ISBETTER", CPUVEN_AMD }           ; Vendor - AMD Prototype
                VENDOR { "AuthenticAMD", CPUVEN_AMD }           ; Vendor - AMD
                VENDOR { "CentaurHauls", CPUVEN_CENTAUR }       ; Vendor - Centaur
                VENDOR { "CyrixInstead", CPUVEN_CYRIX }         ; Vendor - Cyrix & IBM
                VENDOR { "GenuineIntel", CPUVEN_INTEL }         ; Vendor - Intel
                VENDOR { "GenuineTMx86", CPUVEN_TRANSMETA }     ; Vendor - Transmeta
                VENDOR { "NexGenDriven", CPUVEN_NEXGEN }        ; Vendor - NexGen
                VENDOR { "RiseRiseRise", CPUVEN_RISE }          ; Vendor - Rise
                VENDOR { "UMC UMC UMC ", CPUVEN_UMC }           ; Vendor - UMC
ThisVendor      VENDOR <>                                       ; Vendor - Unknown


        tclkl   dd      0               ; Tick Count Low
        tclkh   dd      0               ; Tick Count High



CpuX86  SEGMENT USE32 PUBLIC 'CODE'



;-------------------------------------------------------------------------------------------------
;       void CpuDetect( CpuInfo * const pInfo );
;
;       CPU detection code.  Used similar to that as in the Intel AP-485 document
;       "Intel Processor Identification and the CPUID Instruction" since no
;       real need to change the primary function of the code!
;
;       Win95 uses a minimum of a 80386 and so we start checking for that 
;       processor first!  Although the 286 can handle a form of protected mode we 
;       pretty much ignore it and so a long time ago I ripped the 8086..80286 logic
;       out of this code!
;
;       This code is not optimized in any way as this is only need be called once during 
;       program initialization.

        public  CpuDetectAsm
CpuDetectAsm       proc    near uses ebx esi edi, pInfo:ptr

;       This section tests for a 386 processor by attempting to flip the 'AC' data alignment
;       (fault) bit#18.  Only the 486 and above have this ability.


mov eax,12345678h
;push eax


        sub     esp,8
        and     esp,0FFFFFFF0h
        mov     [esp+0],eax




pop eax

        .386P
        xor     eax,eax
        mov     TheCpuInfo.nMfg,eax             ; Preset Vendor
        mov     TheCpuInfo.nBits,eax            ; Preset no bits
        mov     TheCpuInfo.nCpuId,CPU_386       ; Preset as a 386
        mov     TheCpuInfo.nFpuId,FPU_NONE      ; Preset as no FPU

        pushfd                          ; push EFLAGS register on stack
        pop     eax                     ; pop those flags into eax

        mov     ecx,eax                 ; save a copy of EFLAGS
        xor     eax,EFLAGS_AC           ; flip AC bit#18 in EFLAGS
        push    eax                     ; push modified flags on stack
        popfd                           ; pop those flags back into EFLAGS

        pushfd                          ; Now push resulting EFLAGS on stack
        pop     eax                     ; pop those flags into eax

        xor     eax,ecx                 ; If AC bit didn't flip...
        jz      $FpuTst                 ; then jump, as processor is an 80386

;       This section tests for a 486 processor by attempting to flip the 'Id' CPUID existance
;       bit#21.  Only Pentiums and above have this capablility, except for some models of the
;       486.  With the CPUID instruction we can more easily determine the capabilities and
;       exact design specifications of a processor.
        .486P
        mov     TheCpuInfo.nCpuId,CPU_486       ; Preset as a 486

        mov     eax,ecx                 ; get a copy of original EFLAGS
        xor     eax,EFLAGS_ID           ; flip ID bit#21 in EFLAGS
        push    eax                     ; push modified flags on stack
        popfd                           ; pop those flags back into EFLAGS

        pushfd                          ; Now push resulting EFLAGS on stack
        pop     eax                     ; pop those flags into eax

        xor     eax,ecx                 ; If ID bit didn't flip...
        jz      $FpuTst                 ; then jump, as processor is an 80486!

        push    ecx                     ; Push the original flags
        popfd                           ; Restore EFLAGS back to normal

;       At this point we could be a newer model of the 80486 or a higher level
;       processor.  Life becomes easier because the CPUID will help determine
;       any detailed information we need regarding the chip such as Vendor, Family,
;       Model, Stepping and other features.  For our purposes, we only retain some
;       superficial capablility information.

        .586P
         xor    eax,eax                 ; setup for CPUID instruction
         cpuid                          ; Get and save vendor ID

                ;edx = extended bits (Extended Command#0)

        mov     dword ptr ThisVendor.vname,ebx  ; Save Vendor string
        mov     dword ptr ThisVendor.vname+4,edx
        mov     dword ptr ThisVendor.vname+8,ecx
        mov     eax,- SIZEOF( VENDOR )

;
;       Test for a Processor Name match
;

$vloop: add     eax,SIZEOF(VENDOR)

        cmp     dword ptr VendorList[eax].vname,ebx
        jne     $vloop                  ; Jump if no match

        cmp     dword ptr VendorList[eax].vname+4,edx
        jne     $vloop                  ; Jump if no match

        cmp     dword ptr VendorList[eax].vname+8,ecx
        jne     $vloop                  ; Jump if no match

;               Vendor Record match found

        mov     eax,VendorList[eax].Id  ; Get Vendor Id from table
        mov     TheCpuInfo.nMfg,eax     ; Save it!

;               CPUID feature #1

        mov     eax,1
        cpuid                           ; Get family/model/stepping features

;               Serial #?

        mov     SerialNum+0,0

        test    edx,CPUIDFLG_SERIAL
        jz      nosn                    ; Jump if no serial#

        mov     SerialNum+8,eax         ; bits 95...64

        mov     eax,3
        cpuid

        mov     SerialNum+4,edx         ; bits 63...32
        mov     SerialNum+0,ecx         ; bits 31...0   

        mov     eax,1
        cpuid                           ; Get family/model/stepping features

nosn:   shr     eax,8                   ; isolate family (al <- ah)
        and     eax,0fh
        mov     TheCpuInfo.nCpuId,eax   ; Save cpu_type with family

;               Floating Point Unit ?

        xor     eax,eax
        test    edx,CPUIDFLG_FPU        ; FPU technology bit #0?
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0

;               If CPUID says we have a floating point, then it's built in!

        mov     ecx,TheCpuInfo.nCpuId
        and     ecx,eax                 ; Use the mask
        mov     TheCpuInfo.nFpuId,ecx   ; Save CPU as FPU!

;               Now use the mask for our FPU bit!

        and     eax,CPUBITS_FPU
        or      TheCpuInfo.nBits,eax    ; Flag as having a FP Chip

;               MMX ?

        xor     eax,eax
        test    edx,CPUIDFLG_MMX        ; MMX technology bit #23?
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_MMX
        or      TheCpuInfo.nBits,eax

;               FXSR ?

        xor     eax,eax
        test    edx,CPUIDFLG_FXSR       ; FXSR technology bit #24?
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_FXSR
        or      TheCpuInfo.nBits,eax

;               SSE ?

        xor     eax,eax
        test    edx,CPUIDFLG_SSE        ; SSE technology bit #25?
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_SSE
        or      TheCpuInfo.nBits,eax

;               SSE2 ?

        xor     eax,eax
        test    edx,CPUIDFLG_SSE2       ; MMX technology bit #26?
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_SSE2
        or      TheCpuInfo.nBits,eax


;               Special testing

        mov     eax,TheCpuInfo.nMfg
        jmp     near ptr $CpuSpecialTbl[eax*4]



;       align   16
$CpuUnknown:

$FpuTst: cmp    TheCpuInfo.nFpuId,FPU_NONE      ; If FPU was found to exist
        jne     $Wrapup                         ; Jump to wrap it up

;       At this point, we know that the processor that gets here is an 80486 or below and
;       may not support the CPUID instruction.  We therefore have to work to generate any
;       usable information about the processor.
;
;       Floating Point test for non CPUID able processors

        mov     word ptr Fpu,05a5ah     ; Random (non-zero) value
        fninit                          ; Initialize FP Chip (no wait)
        fnstsw  word ptr Fpu            ; Store FPU Status Word
        cmp     byte ptr Fpu,0          ; Are lower 8 bits set to 0?
        jne     $Wrapup                 ; Jump if no, since no FPU

        fnstcw  word ptr Fpu            ; Store FPU Control Word

        mov     ebx,Fpu                 ; Get the control word
        and     ebx,103fh               ; mask control bits
        cmp     ebx,3fh                 ; Is NPX installed?
        jne     $Wrapup                 ; Jump if no, since no FPU

;               We have a Floating Point Chip so flag it!

        or      TheCpuInfo.nBits,CPUBITS_FPU    ; Flag as having a FP Chip

;               We need a little additional information as the older FPU's
;               had their own unique behaviors.  And some instructions retired!

;       Test CPU 386 for FPU 287 or 387?

        mov     ebx,TheCpuInfo.nCpuId   ; Get CPU type
        cmp     ebx,CPU_386
        ja      $FTst1

        mov     ebx,FPU_287             ; Set as a 80287 FPU
        fld1                            ; Push (+1.0) Constant onto FPU stack
        fldz                            ; Push (0.0) Constant onto FPU stack
        fdiv                            ; Divide
        fld     st                      ; Push ST onto FPU stack
;                                       ; Note: fchs won't invert an invalid # on >=387
        fchs                            ; ST = -(ST)  (Invert sign)
        fcompp                          ; Compare ST(0) with ST(1) & pop twice
        fstsw   ax                      ; Store Status Word

        shr     ax,1                    ; Carry 0=387, 1=287
        sbb     ebx,0ffffffffh          ; Subtract with carry to actually add!
                                        ; +1=387 +0=287

;       If we're here then we are a 486 so we have either a 387 or we
;       have an internal FPU.  Since the 387 and 487 are functionally the same
;       then assign the FPU as the CPU!

$FTst1: mov     TheCpuInfo.nFpuId,ebx   ; Save FpuType

;
;       We're all done except for returning the CPU information!
;
        .386P

$Wrapup:
        mov     ebx,pInfo

;               Return CPU Enhancement information

        mov     eax,TheCpuInfo.nMfg
        mov     (CpuInfo ptr [ebx]).nMfg,eax    ; Set Vendor pointer with Vendor Id

        mov     eax,TheCpuInfo.nBits
        mov     (CpuInfo ptr [ebx]).nBits,eax   ; Save configuration bits

        mov     eax,TheCpuInfo.nCpuId           ; Return CPU type
        mov     (CpuInfo ptr [ebx]).nCpuId,eax

        mov     eax,TheCpuInfo.nFpuId           ; Return FPU type
        mov     (CpuInfo ptr [ebx]).nFpuId,eax
        ret


        .686P
;
;       Extra AMD Mfg. Tests
;
         align  16

$CpuAmd: mov    eax,AMD_EFUNC_0         ; Extended function #1
        cpuid                           ; Test 3D Now bit!

        cmp     eax,AMD_EFUNC_0
        jbe     $Xit                    ; Extended functions NOT supported

;               Do function #1

        mov     eax,AMD_EFUNC_1         ; Extended function #1
        cpuid                           ; Test 3D Now bit!

        mov     ecx,TheCpuInfo.nBits    ; Blend bits

;               3DNow instructions

        xor     eax,eax
        test    edx,AMD_EFLG_3DNOW
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_3DNOW
        or      ecx,eax                 ; Set bit if 3DNow supported

;               3DNow (Extensions) instructions

        xor     eax,eax
        test    edx,AMD_EFLG_3DNOWX
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_3DNOW_EXT
        or      ecx,eax                 ; Set bit if 3DNow Extensions supported

;               3DNow (MMX Extensions) instructions

        xor     eax,eax
        test    edx,AMD_EFLG_MMXEXT
        sete    al                      ;  0  1
        dec     eax                     ; -1, 0
        and     eax,CPUBITS_3DNOW_MMX
        or      ecx,eax                 ; Set bit if 3DNow MMX Extensions supported

;               3DNow! Professional (3DNow! & SSE) Hybrid

        mov     eax,ecx                 ; Get bits
        and     eax,CPUBITS_3DNOW + CPUBITS_SSE
        cmp     eax,CPUBITS_3DNOW + CPUBITS_SSE
        setne   al                      ; 1=NO 0=Yes
        and     eax,00001b
        dec     eax                     ; 0=NO -1=Yes
        and     eax,CPUBITS_3DNOW_SSE
        or      ecx,eax                 ; Set bit if 3DNow! & SSE thus 3DNow! Professional

        or      TheCpuInfo.nBits,ecx    ; Blend bits

$Xit:   jmp     $CpuUnknown



        align   16
;       Jump table for extra (special) feature handling

$CpuSpecialTbl:
        dd      offset $CpuUnknown      ;0 Unknown
        dd      offset $CpuUnknown      ;1 Intel
        dd      offset $CpuAmd          ;2 AMD
        dd      offset $CpuUnknown      ;3 Cyrix
        dd      offset $CpuUnknown      ;4 Centaur
        dd      offset $CpuUnknown      ;5 National
        dd      offset $CpuUnknown      ;6 UMC
        dd      offset $CpuUnknown      ;7 NexGen
        dd      offset $CpuUnknown      ;8 Rise
        dd      offset $CpuUnknown      ;9 Transmeta

CpuDetectAsm    endp



        .K3D
; void vmp_SIMDEntry( void )
;
;       Call upon entry to 3DNow instructions
        align   16
        public  vmp_SIMDEntryAsm3DNow
vmp_SIMDEntryAsm3DNow    proc    near
        femms
        ret
vmp_SIMDEntryAsm3DNow endp


        .MMX
        public  vmp_SIMDEntryAsmMMX
vmp_SIMDEntryAsmMMX    proc    near
        emms
        ret
vmp_SIMDEntryAsmMMX endp




; void vmp_SIMDExit( void )
;
;       Call upon exit (done) with 3DNow instructions
        .K3D
        public  vmp_SIMDExitAsm3DNow
vmp_SIMDExitAsm3DNow    proc    near
        femms
        ret
vmp_SIMDExitAsm3DNow    endp

        .MMX
        public  vmp_SIMDExitAsmMMX
vmp_SIMDExitAsmMMX    proc    near
        emms
        ret
vmp_SIMDExitAsmMMX      endp



;        void CpuDelaySet( void )
;
;        Get&Set current cpu clock cycle
;
;       This function is used to do some simple timing analysis
;        on the duration of intervals.  Call this function to
;        set the initial time then call tclktst() to get the
;        elapsed interval.
;
        align   16

        public  CpuDelaySet
CpuDelaySet     proc    near
         rdtsc                       ; Read time-stamp counter

         mov      tclkl,eax          ; Save low 32bits
         mov      tclkh,edx          ; Save high 32bits
         ret
CpuDelaySet     endp



;        long int CpuDelayCalc( void )
;
;        This function is called after IClkSet() to get the
;        elapsed interval in clock cycles.
;
;        Note: On a 400MHz computer only reading the lower 32bits
;        gives a maximum 10second sampling before rollover.

         align    16
         public   CpuDelayCalc
CpuDelayCalc proc     near
         rdtsc                       ; Read time-stamp counter

         sub      eax,tclkl
         sbb      edx,tclkh          ; edx:eax = total elapsed interval

         ret                         ; return edx:eax = 64bits of info.
CpuDelayCalc endp



CpuX86  ends
        end

