        TITLE   Book PAddX86.asm - Packed Addition
        PAGE    53,160

;
; PAddX86.asm - 
;
; This document contains file that you may use at your own risk in your own programs.
;
; This module is designed to accomodate X86 processors in the book.
;
;       NOTE: Needs masm 6.14 or later!
;
;       Created - 31 Dec 2001 - J.Leiterman
;
;       Modifications
;
;               
;
;       Tabs = 8



        .XLIST

        .387
        .686P
        .MMX
        .model flat, C

        include x86.inc
        include KariType.inc
 
        .LIST



        .data



PAddX86  SEGMENT USE32 PUBLIC 'CODE'


;_________________________________________________________________________________________________
;       (Quad) Vector 16x8bit Summation (Addition)
;
;void vmp_paddB( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
;{
;       *(pbD+0) = *(pbA+0) + *(pbB+0);
;       *(pbD+1) = *(pbA+1) + *(pbB+1);
;       *(pbD+2) = *(pbA+2) + *(pbB+2);
;       *(pbD+3) = *(pbA+3) + *(pbB+3);
;       *(pbD+4) = *(pbA+4) + *(pbB+4);
;       *(pbD+5) = *(pbA+5) + *(pbB+5);
;       *(pbD+6) = *(pbA+6) + *(pbB+6);
;       *(pbD+7) = *(pbA+7) + *(pbB+7);
;       *(pbD+8) = *(pbA+8) + *(pbB+8);
;       *(pbD+9) = *(pbA+9) + *(pbB+9);
;       *(pbD+10) = *(pbA+10) + *(pbB+10);
;       *(pbD+11) = *(pbA+11) + *(pbB+11);
;       *(pbD+12) = *(pbA+12) + *(pbB+12);
;       *(pbD+13) = *(pbA+13) + *(pbB+13);
;       *(pbD+14) = *(pbA+14) + *(pbB+14);
;       *(pbD+15) = *(pbA+15) + *(pbB+15);
;}                                                                      

        align   16
        .MMX
        public  vmp_paddBAsmMMX
vmp_paddBAsmMMX       proc    near pbD:ptr, pbA:ptr, pbB:ptr
;;;        emms
ASSERT_PTR4(pbD)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A
        mov     ecx,pbD                         ; Vector Destination

        movq    mm0,[edx+0]                     ; Read B Data
        movq    mm1,[edx+8]
        movq    mm2,[eax+0]                     ; Read A Data
        movq    mm3,[eax+8]

        paddb   mm0,mm2                         ;       {Al+Bl}
        paddb   mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ;       {Al+Bl}
        movq    [ecx+8],mm1                     ;       {Au+Bu}
        
;;;        emms
        ret
vmp_paddBAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddBAsmSSE20
vmp_paddBAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddb   xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddBAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddBAsmSSE2
vmp_paddBAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddb   xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddBAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed (un)signed Addition 8x16bit
;
;void vmp_paddH( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
;{
;       *(phD+0) = *(phA+0) + *(phB+0);
;       *(phD+1) = *(phA+1) + *(phB+1);
;       *(phD+2) = *(phA+2) + *(phB+2);
;       *(phD+3) = *(phA+3) + *(phB+3);
;       *(phD+4) = *(phA+4) + *(phB+4);
;       *(phD+5) = *(phA+5) + *(phB+5);
;       *(phD+6) = *(phA+6) + *(phB+6);
;       *(phD+7) = *(phA+7) + *(phB+7);
;}                                                                      

        align   16
        .MMX
        public  vmp_paddHAsmMMX
vmp_paddHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phB)
ASSERT_PTR4(phA)
ASSERT_PTR4(phD)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ;vB[0...3] {b3 b2 b1 b0}
        movq    mm1,[edx+8]                     ;vB[4...7] {b7 b6 b5 b4}
        movq    mm2,[eax+0]                     ;vA[0...3] {a3 a2 a1 a0}
        movq    mm3,[eax+8]                     ;vA[4...7] {a7 a6 a5 a4}

        paddw   mm0,mm2                         ;       {Al+Bl}
        paddw   mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ; Write Destination
        movq    [ecx+8],mm1
        
;;;        emms
        ret
vmp_paddHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddHAsmSSE20
vmp_paddHAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddw   xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddHAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddHAsmSSE2
vmp_paddHAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddw   xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed (un)signed Addition 4x32bit
;
;void vmp_paddW( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
;{
;       *(pwD+0) = *(pwA+0) + *(pwB+0);
;       *(pwD+1) = *(pwA+1) + *(pwB+1);
;       *(pwD+2) = *(pwA+2) + *(pwB+2);
;       *(pwD+3) = *(pwA+3) + *(pwB+3);
;}                                                                      

        align   16
        public  vmp_paddWAsmMMX
vmp_paddWAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(phB)
ASSERT_PTR4(phA)
ASSERT_PTR4(phD)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ;vB[0...1] {b1 b0}
        movq    mm1,[edx+8]                     ;vB[2...3] {b3 b2}
        movq    mm2,[eax+0]                     ;vA[0...1] {a1 a0}
        movq    mm3,[eax+8]                     ;vA[2...3] {a3 a2}

        paddd   mm0,mm2                         ;       {Al+Bl}
        paddd   mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ;       {Al+Bl}
        movq    [ecx+8],mm1                     ;       {Au+Bu}
        
;;;        emms
        ret
vmp_paddWAsmMMX endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddWAsmSSE20
vmp_paddWAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddd   xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddWAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddWAsmSSE2
vmp_paddWAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddd   xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddWAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Signed 16x8 Summation (Addition) w/Saturation  {-128...127}

        align   16
        public  vmp_paddsBAsmMMX
vmp_paddsBAsmMMX       proc    near , pbD:ptr, pbA:ptr, pbB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A
        mov     ecx,pbD                         ; Vector Destination

        movq    mm0,[edx+0]                     ;vB[0...7] {b7 b6 b5 b4 b3 b2 b1 b0}
        movq    mm1,[edx+8]                     ;vB[8...15] {b15 b14 b13 b12 b11 b10 b9 b8}
        movq    mm2,[eax+0]                     ;vA[0...7] {a7 a6 a5 a4 a3 a2 a1 a0}
        movq    mm3,[eax+8]                     ;vA[8...15] {a15 a14 a13 a12 a11 a10 a9 a8}

        paddsb  mm0,mm2                         ;       {Al+Bl}
        paddsb  mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ;       {Al+Bl}
        movq    [ecx+8],mm1                     ;       {Au+Bu}
        
;;;        emms
        ret
vmp_paddsBAsmMMX endp




;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddsBAsmSSE20
vmp_paddsBAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddsb  xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddsBAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddsBAsmSSE2
vmp_paddsBAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddsb  xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddsBAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Unsigned 16x8 Summation (Addition) w/Saturation  {0...255}

        align   16
        public  vmp_paddusBAsmMMX
vmp_paddusBAsmMMX       proc    near pbD:ptr, pbA:ptr, pbB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A
        mov     ecx,pbD                         ; Vector Destination

        movq    mm0,[edx+0]                     ;vB[0...7] {b7 b6 b5 b4 b3 b2 b1 b0}
        movq    mm1,[edx+8]                     ;vB[8...15] {b15 b14 b13 b12 b11 b10 b9 b8}
        movq    mm2,[eax+0]                     ;vA[0...7] {a7 a6 a5 a4 a3 a2 a1 a0}
        movq    mm3,[eax+8]                     ;vA[8...15] {a15 a14 a13 a12 a11 a10 a9 a8}

        paddusb mm0,mm2                         ;       {Al+Bl}
        paddusb mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ;       {Al+Bl}
        movq    [ecx+8],mm1                     ;       {Au+Bu}
        
;;;        emms
        ret
vmp_paddusBAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddusBAsmSSE20
vmp_paddusBAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddusb xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddusBAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddusBAsmSSE2
vmp_paddusBAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddusb xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddusBAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Signed 8x16 Summation (Addition) w/Saturation {-32768...32767}
;
        align   16
        public  vmp_paddsHAsmMMX
vmp_paddsHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(phB)
ASSERT_PTR4(phA)
ASSERT_PTR4(phD)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ;vB[0...3] {b3 b2 b1 b0}
        movq    mm1,[edx+8]                     ;vB[4...7] {b7 b6 b5 b4}
        movq    mm2,[eax+0]                     ;vA[0...3] {a3 a2 a1 a0}
        movq    mm3,[eax+8]                     ;vA[4...7] {a7 a6 a5 a4}

        paddsw  mm0,mm2                         ;       {Al+Bl}
        paddsw  mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ;       {Al+Bl}
        movq    [ecx+8],mm1                     ;       {Au+Bu}
        
;;;        emms
        ret
vmp_paddsHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddsHAsmSSE20
vmp_paddsHAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddsw  xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddsHAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddsHAsmSSE2
vmp_paddsHAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddsw  xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddsHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Unsigned 8x16 Summation (Addition) w/Saturation {0...65535}
;
        align   16
        public  vmp_paddusHAsmMMX
vmp_paddusHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(phB)
ASSERT_PTR4(phA)
ASSERT_PTR4(phD)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ;vB[0...3] {b3 b2 b1 b0}
        movq    mm1,[edx+8]                     ;vB[4...7] {b7 b6 b5 b4}
        movq    mm2,[eax+0]                     ;vA[0...3] {a3 a2 a1 a0}
        movq    mm3,[eax+8]                     ;vA[4...7] {a7 a6 a5 a4}

        paddusw mm0,mm2                         ;       {Al+Bl}
        paddusw mm1,mm3                         ;       {Au+Bu}

        movq    [ecx+0],mm0                     ;       {Al+Bl}
        movq    [ecx+8],mm1                     ;       {Au+Bu}
        
;;;        emms
        ret
vmp_paddusHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_paddusHAsmSSE20
vmp_paddusHAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbB)
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddusw xmm0,xmm1                       ;       {vA+vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_paddusHAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_paddusHAsmSSE2
vmp_paddusHAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbB)
ASSERT_PTR16(pbA)
ASSERT_PTR16(pbD)

        mov     edx,pbB                         ; Vector B
        mov     eax,pbA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,pbD                         ; Vector Destination
        paddusw xmm0,xmm1                       ;       {vA+vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_paddusHAsmSSE2 endp




PAddX86  ends
        end

