
//
//  PasX86.cpp - Packed Addition/Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "Pas.h"
#include "PasEx.h"



//
//  Setup Packed Addition/Subtraction
//
//  NOTE: This function sets up the function vectors so the best function
//      is attached to the vector for the processor running.
//

void PasSetup( const CpuInfo * const pcinfo )
{
    ASSERT_PTR4(pcinfo);
        
    if ( CPUBITS_SSE2 & pcinfo->nBits )
      {
        vmp_paddB =             vmp_paddBAsmSSE2;
        vmp_paddB0 =            vmp_paddBAsmSSE20;
        vmp_paddH =             vmp_paddHAsmSSE2;
        vmp_paddH0 =            vmp_paddHAsmSSE20;
        vmp_paddW =             vmp_paddWAsmSSE2;
        vmp_paddW0 =            vmp_paddWAsmSSE20;

        vmp_psubB =             vmp_psubBAsmSSE2;
        vmp_psubB0 =            vmp_psubBAsmSSE20;
        vmp_psubH =             vmp_psubHAsmSSE2;
        vmp_psubH0 =            vmp_psubHAsmSSE20;
        vmp_psubW =             vmp_psubWAsmSSE2;
        vmp_psubW0 =            vmp_psubWAsmSSE20;

        vmp_paddusB =           vmp_paddusBAsmSSE2;
        vmp_paddusB0 =          vmp_paddusBAsmSSE20;
        vmp_paddusH =           vmp_paddusHAsmSSE2;
        vmp_paddusH0 =          vmp_paddusHAsmSSE20;

        vmp_psubusB =           vmp_psubusBAsmSSE2;
        vmp_psubusB0 =          vmp_psubusBAsmSSE20;
        vmp_psubusH =           vmp_psubusHAsmSSE2;
        vmp_psubusH0 =          vmp_psubusHAsmSSE20;

        vmp_paddsB =            vmp_paddsBAsmSSE2;
        vmp_paddsB0 =           vmp_paddsBAsmSSE20;
        vmp_paddsH =            vmp_paddsHAsmSSE2;
        vmp_paddsH0 =           vmp_paddsHAsmSSE20;

        vmp_psubsB =            vmp_psubsBAsmSSE2;
        vmp_psubsB0 =           vmp_psubsBAsmSSE20;
        vmp_psubsH =            vmp_psubsHAsmSSE2;
        vmp_psubsH0 =           vmp_psubsHAsmSSE20;
      }

    else if ( CPUBITS_MMX & pcinfo->nBits )
      {
        vmp_paddB =             vmp_paddBAsmMMX;
        vmp_paddB0 =            vmp_paddBAsmMMX;
        vmp_paddH =             vmp_paddHAsmMMX;
        vmp_paddH0 =            vmp_paddHAsmMMX;
        vmp_paddW =             vmp_paddWAsmMMX;
        vmp_paddW0 =            vmp_paddWAsmMMX;

        vmp_psubB =             vmp_psubBAsmMMX;
        vmp_psubB0 =            vmp_psubBAsmMMX;
        vmp_psubH =             vmp_psubHAsmMMX;
        vmp_psubH0 =            vmp_psubHAsmMMX;
        vmp_psubW =             vmp_psubWAsmMMX;
        vmp_psubW0 =            vmp_psubWAsmMMX;

        vmp_paddusB =           vmp_paddusBAsmMMX;
        vmp_paddusB0 =          vmp_paddusBAsmMMX;
        vmp_paddusH =           vmp_paddusHAsmMMX;
        vmp_paddusH0 =          vmp_paddusHAsmMMX;

        vmp_psubusB =           vmp_psubusBAsmMMX;
        vmp_psubusB0 =          vmp_psubusBAsmMMX;
        vmp_psubusH =           vmp_psubusHAsmMMX;
        vmp_psubusH0 =          vmp_psubusHAsmMMX;

        vmp_paddsB =            vmp_paddsBAsmMMX;
        vmp_paddsB0 =           vmp_paddsBAsmMMX;
        vmp_paddsH =            vmp_paddsHAsmMMX;
        vmp_paddsH0 =           vmp_paddsHAsmMMX;
                                          
        vmp_psubsB =            vmp_psubsBAsmMMX;
        vmp_psubsB0 =           vmp_psubsBAsmMMX;
        vmp_psubsH =            vmp_psubsHAsmMMX;
        vmp_psubsH0 =           vmp_psubsHAsmMMX;
      }

    else
      {
        vmp_paddB =             vmp_paddBGeneric;
        vmp_paddB0 =            vmp_paddBGeneric;
        vmp_paddH =             vmp_paddHGeneric;
        vmp_paddH0 =            vmp_paddHGeneric;
        vmp_paddW =             vmp_paddWGeneric;
        vmp_paddW0 =            vmp_paddWGeneric;

        vmp_psubB =             vmp_psubBGeneric;
        vmp_psubB0 =            vmp_psubBGeneric;
        vmp_psubH =             vmp_psubHGeneric;
        vmp_psubH0 =            vmp_psubHGeneric;
        vmp_psubW =             vmp_psubWGeneric;
        vmp_psubW0 =            vmp_psubWGeneric;

        vmp_paddusB =           vmp_paddusBGeneric;
        vmp_paddusB0 =          vmp_paddusBGeneric;
        vmp_paddusH =           vmp_paddusHGeneric;
        vmp_paddusH0 =          vmp_paddusHGeneric;

        vmp_psubusB =           vmp_psubusBGeneric;
        vmp_psubusB0 =          vmp_psubusBGeneric;
        vmp_psubusH =           vmp_psubusHGeneric;
        vmp_psubusH0 =          vmp_psubusHGeneric;

        vmp_paddsB =            vmp_paddsBGeneric;
        vmp_paddsB0 =           vmp_paddsBGeneric;
        vmp_paddsH =            vmp_paddsHGeneric;
        vmp_paddsH0 =           vmp_paddsHGeneric;

        vmp_psubsB =            vmp_psubsBGeneric;
        vmp_psubsB0 =           vmp_psubsBGeneric;
        vmp_psubsH =            vmp_psubsHGeneric;
        vmp_psubsH0 =           vmp_psubsHGeneric;
      }
}




