        TITLE   QVas3DX86.asm - 80(x)86 3D Math
        PAGE    53,160

;  QVas3D.asm - 3D (Quad) Math Addition & Subtraction
;
;  This document contains file that you may use at your own risk in your own programs.
;
;  This module is designed to accomodate the book.
;
;               This module is designed for fast 3D math
;
;  Created - 3 Oct 2001 - James Leiterman
;
;  Last Modification
;
;       Tabs = 8



        .XLIST

        .686
        .XMM

        .model flat, C


        include KariType.inc
        include vmp3DX86M.inc
 
        .LIST



        .data
        align   16



;------------------------------------------------------------------------------------------------


QVas3DX86       SEGMENT USE32 PUBLIC 'CODE'

;-------------------------------------------------------------------------------------------------



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Summation (Addition)
;
;void vmp_QVecAdd(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
;{
;       pvD->x = pvA->x + pvB->x;
;       pvD->y = pvA->y + pvB->y;
;       pvD->z = pvA->z + pvB->z;
;       pvD->w = pvA->w + pvB->w;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DQVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_QVecAddAsm3DNow
vmp_QVecAddAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr
        .K3D
;;;        femms
ASSERT_PTR4(vD)
ASSERT_PTR4(vA)
ASSERT_PTR4(vB)

        mov     ecx,vA                          ; Vector A
        mov     edx,vB                          ; Vector B
        mov     eax,vD                          ; Vector Destination

        movq    mm0,[ecx+0]                     ;vA.xy   {Ay Ax}
        movq    mm2,[edx+0]                     ;vB.xy   {By Bx}
        movq    mm1,[ecx+8]                     ;vA.zw   {Aw Az}
        movq    mm3,[edx+8]                     ;vB.zw   {Bw Bz}

        pfadd   mm0,mm2                         ;       {Ay+By Ax+Bx}
        pfadd   mm1,mm3                         ;       {Aw+Bw Az+Bz}

        movq    [eax+0],mm0                     ;       {Ay+By Ax+Bx}
        movq    [eax+8],mm1                     ;       {Aw+Bw Az+Bz}
        
;;;        femms
        ret
vmp_QVecAddAsm3DNow endp


;_________________________________________________________________________________________________

; *** SSE version *** non aligned
;
        align   16
        public  vmp_QVecAddAsmSSE0
vmp_QVecAddAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR4(vB)
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movups  xmm1,[edx]                      ;vB.xyzw {Bw Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyzw {Aw Az Ay Ax}

        mov     eax,vD                          ; Vector Destination
        addps   xmm0,xmm1                       ;        {Aw+Bw Az+Bz Ay+By Ax+Bx}
        movups  [eax],xmm0                      ;        {Aw+Bw Az+Bz Ay+By Ax+Bx}      
        
;;;        emms
        ret
vmp_QVecAddAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_QVecAddAsmSSE
vmp_QVecAddAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR16(vB)
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movaps  xmm1,[edx]                      ;vB.xyz# {Bw Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyz# {Aw Az Ay Ax}

        mov     eax,vD                          ; Vector Destination
        addps   xmm0,xmm1                       ;        {Aw+Bw Az+Bz Ay+By Ax+Bx}
        movaps  [eax],xmm0                      ;        {Aw+Bw Az+Bz Ay+By Ax+Bx}      
        
;;;        emms
        ret
vmp_QVecAddAsmSSE endp


;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Scalar Summation (Addition)
;
;void vmp_QVecAddScalar(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
;{
;       pvD->x = pvA->x + fScalar;
;       pvD->y = pvA->y + fScalar;
;       pvD->z = pvA->z + fScalar;
;       pvD->w = pvA->w + fScalar;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_QVecAddScalarAsm3DNow
vmp_QVecAddScalarAsm3DNow       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .K3D
;;;        femms
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        movd    mm2,fScalar                     ; fScalar {0 s}
        mov     ecx,vA                          ; Vector A
        punpckldq mm2,mm2                               ; fScalar {s s}
        
        mov     eax,vD                          ; Vector Destination

        movq    mm0,[ecx+0]                     ;vA.xy   {Ay Ax}
        movq    mm1,[ecx+8]                     ;vA.zw   {Aw Az}

        pfadd   mm0,mm2                         ;       {Ay+s  Ax+s}
        pfadd   mm1,mm2                         ;       {Aw+s  Az+s}

        movq    [eax+0],mm0                     ;       {Ay+By Ax+Bx}
        movq    [eax+8],mm1                     ;       {Aw+Bw Az+Bz}
        
;;;        femms
        ret
vmp_QVecAddScalarAsm3DNow endp


;_________________________________________________________________________________________________

; *** SSE version *** non aligned
;
        align   16
        public  vmp_QVecAddScalarAsmSSE0
vmp_QVecAddScalarAsmSSE0       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        mov     ecx,vA                          ; Vector A
        movss   xmm1,fScalar                    ;         {0  0  0  s}

        movups  xmm0,[ecx]                      ;vA.xyzw  {Aw Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0 {s  s  s  s}

        mov     eax,vD                          ; Vector Destination
        addps   xmm0,xmm1                       ;        {Aw+s Az+s Ay+s Ax+s}

        movups  [eax],xmm0                      ;        {Aw+s Az+s Ay+s Ax+s}  
        
;;;        emms
        ret
vmp_QVecAddScalarAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_QVecAddScalarAsmSSE
vmp_QVecAddScalarAsmSSE       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

        mov     ecx,vA                          ; Vector A
        movss   xmm1,fScalar                    ;         {0  0  0  s}

        movaps  xmm0,[ecx]                      ;vA.xyzw  {Aw Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0 {s  s  s  s}

        mov     eax,vD                          ; Vector Destination
        addps   xmm0,xmm1                       ;        {Aw+s Az+s Ay+s Ax+s}

        movaps  [eax],xmm0                      ;        {Aw+s Az+s Ay+s Ax+s}  
        
;;;        emms
        ret
vmp_QVecAddScalarAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Difference (Subtraction)
;
;void vmp_QVecSub(vmp3DVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
;{
;       pvD->x = pvA->x - pvB->x;
;       pvD->y = pvA->y - pvB->y;
;       pvD->z = pvA->z - pvB->z;
;       pvD->w = pvA->w - pvB->w;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_QVecSubAsm3DNow
vmp_QVecSubAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr
        .K3D
;;;        femms
ASSERT_PTR4(vB)
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        mov     ecx,vA                          ; Vector A
        mov     edx,vB                          ; Vector B
        mov     eax,vD                          ; Vector Destination

        movq    mm0,[ecx+0]                     ;vA.xy   {Ay Ax}
        movq    mm2,[edx+0]                     ;vB.xy   {By Bx}
        movq    mm1,[ecx+8]                     ;vA.zw   {Aw Az}
        movq    mm3,[edx+8]                     ;vB.zw   {Bw Bz}

        pfsub   mm0,mm2                         ;       {Ay-By Ax-Bx}
        pfsub   mm1,mm3                         ;       {Aw-Bw Az-Bz}

        movq    [eax+0],mm0                     ;       {Ay-By Ax-Bx}
        movq    [eax+8],mm1                     ;       {Aw-Bw Az-Bz}
        
;;;        femms
        ret
vmp_QVecSubAsm3DNow endp


;_________________________________________________________________________________________________

; *** SSE version *** non aligned
;
        align   16
        public  vmp_QVecSubAsmSSE0
vmp_QVecSubAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR4(vB)
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movups  xmm1,[edx]                      ;vB.xyzw {Bw Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyzw {Aw Az Ay Ax}

        mov     eax,vD                          ; Vector Destination
        subps   xmm0,xmm1                       ;        {Aw-Bw Az-Bz Ay-By Ax-Bx}
        movups  [eax],xmm0                      ;        {Aw-Bw Az-Bz Ay-By Ax-Bx}      
        
;;;        emms
        ret
vmp_QVecSubAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_QVecSubAsmSSE
vmp_QVecSubAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR16(vB)
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movaps  xmm1,[edx]                      ;vB.xyzw {Bw Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyzw {Aw Az Ay Ax}

        mov     eax,vD                          ; Vector Destination
        subps   xmm0,xmm1                       ;        {Aw-Bw Az-Bz Ay-By Ax-Bx}
        movaps  [eax],xmm0                      ;        {Aw-Bw Az-Bz Ay-By Ax-Bx}      
        
;;;        emms
        ret
vmp_QVecSubAsmSSE endp


;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Scalar Difference (Subtraction)
;
;void vmp_QVecSubScalar(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
;{
;       pvD->x = pvA->x - fScalar;
;       pvD->y = pvA->y - fScalar;
;       pvD->z = pvA->z - fScalar;
;       pvD->w = pvA->w - fScalar;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_QVecSubScalarAsm3DNow
vmp_QVecSubScalarAsm3DNow       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .K3D
;;;        femms
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        movd    mm2,fScalar                     ; fScalar {0 s}
        mov     ecx,vA                          ; Vector A
        punpckldq mm2,mm2                               ;         {s s}
        
        mov     eax,vD                          ; Vector Destination

        movq    mm0,[ecx+0]                     ;vA.xy   {Ay Ax}
        movq    mm1,[ecx+8]                     ;vA.zw   {Aw Az}

        pfsub   mm0,mm2                         ;       {Ay-s Ax-s}
        pfsub   mm1,mm2                         ;       {Aw-s Az-s}

        movq    [eax+0],mm0                     ;       {Ay-By Ax-Bx}
        movq    [eax+8],mm1                     ;       {Aw-Bw Az-Bz}
        
;;;        femms
        ret
vmp_QVecSubScalarAsm3DNow endp


;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_QVecSubScalarAsmSSE0
vmp_QVecSubScalarAsmSSE0       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

        mov     ecx,vA                          ; Vector A
        movss   xmm1,fScalar                    ;         {0  0  0  s}

        movups  xmm0,[ecx]                      ;vA.xyzw  {Aw Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0 {s  s  s  s}

        mov     eax,vD                          ; Vector Destination
        subps   xmm0,xmm1                       ;        {Aw-s Az-s Ay-s Ax-s}

        movups  [eax],xmm0                      ;        {Aw-s Az-s Ay-s Ax-s}  
        
;;;        emms
        ret
vmp_QVecSubScalarAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_QVecSubScalarAsmSSE
vmp_QVecSubScalarAsmSSE       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

        mov     ecx,vA                          ; Vector A
        movss   xmm1,fScalar                    ;         {0  0  0  s}

        movaps  xmm0,[ecx]                      ;vA.xyzw  {Aw Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0 {s  s  s  s}

        mov     eax,vD                          ; Vector Destination
        subps   xmm0,xmm1                       ;        {Aw-s Az-s Ay-s Ax-s}

        movaps  [eax],xmm0                      ;        {Aw-s Az-s Ay-s Ax-s}  
        
;;;        emms
        ret
vmp_QVecSubScalarAsmSSE endp



QVas3DX86       ends
        end

