
//
//  VTrig3D.cpp - 3D Vector Trignometric Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <iostream.h>
#include    <math.h>

    //  Book Development Library Modules

#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "VTrig3D.h"                                 // 3D Math
#include    "VTrig3DEx.h"



vmp_FSinCosProc         vmp_FSinCos;
vmp_FSinProc            vmp_FSin;
vmp_FCosProc            vmp_FCos;
vmp_FTanProc            vmp_FTan;

vmp_FASinProc           vmp_FASin;
vmp_FACosProc           vmp_FACos;
vmp_FATanProc           vmp_FATan;
vmp_FATan2Proc          vmp_FATan2;




void vmp_FSinCosGeneric( float * const pfSin, float * const pfCos, float fRads )
{
    ASSERT_PTR4( pfSin );
    ASSERT_PTR4( pfCos );

    *pfSin = sinf( fRads );
    *pfCos = cosf( fRads );
}


void vmp_FSinGeneric( float * const pfD, float fRads )
{
    ASSERT_PTR4( pfD );

    *pfD = sinf( fRads );
}


void vmp_FCosGeneric( float * const pfD, float fRads )
{
    ASSERT_PTR4( pfD );

    *pfD = cosf( fRads );
}


void vmp_FTanGeneric( float * const pfD, float fRads )
{
    ASSERT_PTR4( pfD );

    *pfD = tanf( fRads );
}



//
//  ArcSine of X
//
//  returns {-n/2...n/2 radians}

void vmp_FASinGeneric( float * const pfD, float x )
{
    ASSERT_PTR4( pfD );
    ASSERT(1 >= x);
    ASSERT(-1<= x);
 
    *pfD = asinf( x );
}



//
//  ArcCosine of X
//
//  returns {0...n radians}

void vmp_FACosGeneric( float * const pfD, float x )
{
    ASSERT_PTR4( pfD );
    ASSERT(1 >= x);
    ASSERT(-1<= x);
 
    *pfD = acosf( x );
}



//
//  ArcTangent of X
//
//  if x == 0 then returns 0.
//
//  returns {-n/2...n/2 radians}

void vmp_FATanGeneric( float * const pfD, float x )
{
    ASSERT_PTR4( pfD );
 
    *pfD = atanf( x );
}



//
//  ArcTangent of y/x
//
//  if x and y == 0 then returns 0.
//
//  returns {-n...n radians}
//  Use the sign of both y and x to determine quadrant.

void vmp_FATan2Generic( float * const pfD, float y, float x )
{
    ASSERT_PTR4( pfD );
 
    *pfD = atan2f( y, x );
}



//
//

void vmp_FCotGeneric( float * const pfD, float x )
{
    ASSERT_PTR4( pfD );
 
    *pfD = cosf( x ) / sinf( x );
}


//
//

void vmp_FSecGeneric( float * const pfD, float x )
{
    ASSERT_PTR4( pfD );
 
    *pfD = 1 / cosf( x );
}


//
//

void vmp_FCscGeneric( float * const pfD, float x )
{
    ASSERT_PTR4( pfD );
 
    *pfD = 1 / sinf( x );
}



// sin(x+y) = sin(x)cos(y)+sin(y)cos(x)
// cos(x+y)=cos(x)cos(y)-sin(x)sin(y)
// sin(2x)=cos^2x-sin^2x
// 1+tan^2(x)=sec^2(x)
// 1+cot^2(x)=csc^2(x)


