
//
//  Bench.cpp - Vector 3D Quaternions Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      Hopefully by now you have realized the necessity of 16 byte aligned data
//      and so this program is written as such!
//
//      Because of all the Generic code there is an excessive amount of SIMD entry/exit
//      function calls. This gets dramatically reduced by
//
//  Uses: Chap03\cpuid
//        Chap08\qVas3d
//        Chap12\matrix3d
//        Chap18\vecBug
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <iostream.h>
#include    <memory.h>
#include <D3dx8math.h>

    //  Book Development Library Modules

#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "QVas3D.h"                                     // QVec Add & Sub
#include "QVas3DEx.h"
#include "Matrix3D.h"                                   // Matrix Module
#include "Matrix3DEx.h"
#include "Quat3D.h"                                     // Quaternion Module
#include "Quat3DEx.h"

#include "vecBug.h"



void vmp_VecNormalizeGeneric(vmp3DVector * const pvD, const vmp3DVector * const pvA);


//
//  Vector Normalization
//

void vmp_VecNormalizeGeneric(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
    float fMag;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    fMag = pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z; // + pvA->w * pvA->w;

    if ( fMag < SINGLE_PRECISION )  // If too close to zero...
      {
        pvD->x = pvA->x;
        pvD->y = pvA->y;
        pvD->z = pvA->z;
//      pvD->w = pvA->w;
      }
    else
      {
        fMag = 1.0f / sqrtf(fMag);
        pvD->x = pvA->x * fMag;
        pvD->y = pvA->y * fMag;
        pvD->z = pvA->z * fMag;
//      pvD->w = pvA->w * fMag;
      }
}



//
//  Vector Normalization
//

void vmp_QVecNormalizeGeneric(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
    float fMag;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    fMag = pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z + pvA->w * pvA->w;

    if ( fMag < SINGLE_PRECISION )  // If too close to zero...
      {
        pvD->x = pvA->x;
        pvD->y = pvA->y;
        pvD->z = pvA->z;
      }
    else
      {
        fMag = 1.0f / sqrtf(fMag);
        pvD->x = pvA->x * fMag;
        pvD->y = pvA->y * fMag;
        pvD->z = pvA->z * fMag;
      }

    pvD->w = 0.0f;
}



static void testQuaternions( void )
{
    float matrixbuf[23  *4 +3]; // enough space + 128bit alignment padding
    float fAngX, fAngY, fAngZ, fAngW, fAngV;
    float fAngX2, fAngY2, fAngZ2, fAngW2, fAngV2, fStep;
    float fPrecision;
    float fGAng, fAng;
    vmpQuat *pQTVec, *pQTGVec, *pQTVec1, *pQTGVec1, *pQTVec2, *pQTGVec2, *pQTNorm, *pQTest;
    vmp3DVector *pVec, *pVec0, *pGVec, *pVec2, *pGVec2;
    vmp3DQVector *pQVec, *pQVec0;
    vmp3DMatrix *pmGTmp, *pmTmp;
    vmpQuat qtest={0.0f, 0.0f, 1.0f, 0.0f};

    fPrecision = SINGLE_PRECISION * 10.0f;
    fStep = 30.0f;

        //  Adjust Alignment factor
        //7 Quaternions

    pQTGVec  = (vmpQuat*) ALIGN16((int)matrixbuf);
    pQTVec   = pQTGVec + 1;
    pQTGVec1 = pQTVec + 1;
    pQTVec1  = pQTGVec1 + 1;
    pQTGVec2 = pQTVec1 + 1;
    pQTVec2  = pQTGVec2 + 1;
    pQTNorm  = pQTVec2 + 1;
    pQTest   = pQTNorm + 1;
    
        //2 *4  Matrix
    pmGTmp = (vmp3DMatrix *) (pQTest+1);
    pmTmp  = (vmp3DMatrix *) (pmGTmp+1);

        //2 Vectors (4 float)

    pQVec = (vmp3DQVector *) (pmTmp+1);
    pQVec0 =(vmp3DQVector *) (pQVec+1);

        //5 Vectors (3 float)

    pGVec = (vmp3DVector *) ALIGN16((int)(pQVec0+1));
    pVec =  (vmp3DVector *) ALIGN16((int)(pGVec+1));
    pVec0 = (vmp3DVector *) ALIGN16((int)(pVec+1));
    pGVec2 =(vmp3DVector *) ALIGN16((int)(pVec0+1));
    pVec2 = (vmp3DVector *) ALIGN16((int)(pGVec2+1));

    vmp_QuatCopy(pQTest, &qtest);

        //  ____________________________________________
        //  Rotations Check

    for ( fAngX = -360.0f; fAngX <= 360.0f; fAngX += fStep )    // Yaw
      {
        fAngX2 = DEG2RAD( fAngX );
        cout << "." << flush;

        for ( fAngY = -360.0f; fAngY <= 360.0f; fAngY += fStep )    // Pitch
          {
            fAngY2 = DEG2RAD( fAngY );

            for ( fAngZ = -360.0f; fAngZ <= 360.0f; fAngZ += fStep )    // Roll
              {
                fAngZ2 = DEG2RAD( fAngZ );

                pQVec->x = pVec->x = fAngX;
                pQVec->y = pVec->y = fAngY;
                pQVec->z = pVec->z = fAngZ;

                vmp_VecNormalizeGeneric(pVec0, pVec);

                for ( fAngW = -360.0f; fAngW <= 360.0f; fAngW += fStep )    // Angle
                  {
                    fAngW2 = DEG2RAD( fAngW );

                    pQVec->w = fAngW;
                    vmp_QVecNormalizeGeneric(pQVec0, pQVec);

                        //  Quaternion From Axis Rotation

//                  D3DXQuaternionRotationAxis( (D3DXQUATERNION*)pQTVec, (CONST D3DXVECTOR3 *) pVec0, fAngW2);
                    vmp_QuatRotAxisGeneric( pQTGVec, pVec0, fAngW2 );
                    vmp_QuatRotAxis(        pQTVec,  pVec0, fAngW2 );

                    if ( !vmp_IsQuatEqual( pQTGVec, pQTVec, fPrecision ))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " Quat from AxisRot" << endl; 
                        cout << "    x:" << pQTGVec->x << " y:" << pQTGVec->y << " z:" << pQTGVec->z << " w:" << pQTGVec->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec->x  << " y:" << pQTVec->y  << " z:" << pQTVec->z  << " w:" << pQTVec->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion To Axis Rotation

//                  D3DXQuaternionToAxisAngle( (CONST D3DXQUATERNION*)pQTVec, (D3DXVECTOR3 *) pVec2, &fAng);
                    vmp_QuatToRotAxisGeneric( pGVec2, &fGAng, pQTGVec );
                    vmp_QuatToRotAxis(        pVec2,  &fAng,  pQTGVec );

                    if (!vmp_IsVEqual( pGVec2, pVec2, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Quat to AxisRot" << endl; 
                        cout << "    x:" << pGVec2->x << " y:" << pGVec2->y << " z:" << pGVec2->z << "  Good vs Bad!!!\n";
                        cout << "    x:" << pVec2->x  << " y:" << pVec2->y  << " z:" << pVec2->z  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                    if (!vmp_IsFEqual(fGAng, fAng, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Quat to AxisRot" << endl; 
                        cout << " " << fGAng << " vs " << fAng << "  *** MISMATCH!!! ***" << endl;
                      }

                    D3DXQuaternionNormalize( (D3DXQUATERNION*)pQTNorm, (D3DXQUATERNION*)pQTGVec);

                        //  Quaternion Dot Product

//                  fAng = D3DXQuaternionDot((D3DXQUATERNION*)pQTGVec, (D3DXQUATERNION*)pQTest);
                    vmp_QuatDotProductGeneric(&fGAng, pQTGVec, pQTest);
                    vmp_SIMDEntry();
                    vmp_QuatDotProduct(       &fAng,  pQTGVec, pQTest);
                    vmp_SIMDExit();
                    
                    if (!vmp_IsFEqual(fGAng, fAng, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Dot Product" << endl; 
                        cout << " " << fGAng << " vs " << fAng << "  *** MISMATCH!!! ***" << endl;
                      }

                        //  Quaternion Magnitude

//                  fAng = D3DXQuaternionLength((D3DXQUATERNION*)pQTGVec);
                    vmp_QuatMagnitudeGeneric(&fGAng, pQTGVec);
                    vmp_SIMDEntry();
                    vmp_QuatMagnitude(       &fAng,  pQTGVec);
                    vmp_SIMDExit();
                    
                    if (!vmp_IsFEqual(fGAng, fAng, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Magnitude" << endl; 
                        cout << " " << fGAng << " vs " << fAng << "  *** MISMATCH!!! ***" << endl;
                      }

                        //  Quaternion Magnitude ***FAST***

//                  fAng = D3DXQuaternionLength((D3DXQUATERNION*)pQTGVec);
                    vmp_QuatMagnitudeGeneric(&fGAng, pQTGVec);
                    vmp_SIMDEntry();
                    vmp_QuatMagnitudeFast(       &fAng,  pQTGVec);
                    vmp_SIMDExit();
                    
                    if (!vmp_IsFEqual(fGAng, fAng, FAST_PRECISION))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Magnitude" << endl; 
                        cout << " " << fGAng << " vs " << fAng << "  *** MISMATCH!!! ***" << endl;
                      }

                        //  Quaternion Normalize

                    D3DXQuaternionNormalize( (D3DXQUATERNION*)pQTVec1, (D3DXQUATERNION*)pQTGVec);
//                  vmp_QuatNormalizeGeneric( pQTGVec1, pQTGVec );
                    vmp_SIMDEntry();
//                  vmp_QuatNormalize(        pQTVec1,  pQTGVec );
        vmp_QuatNormalize(        pQTGVec1,  pQTGVec );
                    vmp_SIMDExit();

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Normalize" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Addition

                    vmp_QuatAddGeneric( pQTGVec1, pQTGVec, pQTest );
                    vmp_SIMDEntry();
                    vmp_QuatAdd(        pQTVec1,  pQTGVec, pQTest );
                    vmp_SIMDExit();

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Addition" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Subtraction

                    vmp_QuatSubGeneric( pQTGVec1, pQTGVec, pQTest );
                    vmp_SIMDEntry();
                    vmp_QuatSub(        pQTVec1,  pQTGVec, pQTest );
                    vmp_SIMDExit();

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Subtraction" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Multiplication  |D|=|A|x|B| but D3D is |D|=|B|x|A|

 //                 D3DXQuaternionMultiply( (D3DXQUATERNION*)pQTVec1, (D3DXQUATERNION*)pQTest, (D3DXQUATERNION*)pQTGVec);
                    vmp_QuatMulGeneric( pQTGVec1, pQTGVec, pQTest );
                    vmp_QuatMul(        pQTVec1,  pQTGVec, pQTest );

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Multiplication" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Divide

                    vmp_QuatDivGeneric( pQTGVec1, pQTGVec, pQTest );
                    vmp_QuatDiv(        pQTVec1,  pQTGVec, pQTest );

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Divide" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Conjugate

                    D3DXQuaternionConjugate( (D3DXQUATERNION*)pQTVec1, (D3DXQUATERNION*)pQTGVec);
//                  vmp_QuatConjugateGeneric( pQTGVec1, pQTGVec );
                    vmp_SIMDEntry();
//                  vmp_QuatConjugate(        pQTVec1,  pQTGVec );
                vmp_QuatConjugate(        pQTGVec1,  pQTGVec );
                    vmp_SIMDExit();

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Inverse" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Negate

                    vmp_QuatNegGeneric( pQTGVec1, pQTGVec );
                    vmp_QuatNeg(        pQTVec1,  pQTGVec );

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Negative" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Inverse


                    D3DXQuaternionInverse( (D3DXQUATERNION*)pQTVec1, (D3DXQUATERNION*)pQTGVec);
//                  vmp_QuatInvGeneric( pQTGVec1, pQTGVec );
                    vmp_SIMDEntry();
//                  vmp_QuatInv(        pQTVec1,  pQTGVec );
    vmp_QuatInv(        pQTGVec1,  pQTGVec );
                    vmp_SIMDExit();

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Inverse" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                        //  Quaternion Exponential

//                  D3DXQuaternionExp( (D3DXQUATERNION*)pQTVec1, (D3DXQUATERNION*)pQTNorm);
                    vmp_QuatExpGeneric( pQTGVec1, pQTNorm );
                    vmp_QuatExp(        pQTVec1,  pQTNorm );

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Exponent" << endl; 
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }


                    vmp_SIMDEntry();

                        //  Quaternion Natural Log
#if 01
                    if (0.0f <= pQTNorm->w)
                      {
//                      D3DXQuaternionLn( (D3DXQUATERNION*)pQTVec1, (D3DXQUATERNION*)pQTNorm);

                vmp_SIMDEntry();
                        vmp_QuatLnGeneric( pQTGVec1, pQTNorm );
                        vmp_QuatLn(        pQTVec1,  pQTNorm );

                        if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                          {
                            cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Natural Log" << endl; 
                            cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                            cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                          }
                      }
#endif

                        //  Quaternion to Matrix

                    vmp_SIMDEntry();

//                  D3DXMatrixRotationQuaternion( (D3DXMATRIX*)pmTmp, (D3DXQUATERNION*)pQTNorm);
                    vmp_QuatToMatrixGeneric( *pmGTmp, pQTNorm );
                    vmp_QuatToMatrix(        *pmTmp,  pQTNorm );

                    if ( !vmp_IsMatrixEqual( *pmGTmp, *pmTmp, fPrecision ))
                      {
                        cout << " Quaternion to Matrix diff failure!" << endl;
                        vmp_MatrixPrint(*pmGTmp);
                        vmp_MatrixPrint(*pmTmp);
                      }

                        //  Matrix to Quaternion

                    vmp_SIMDEntry();

//                  D3DXQuaternionRotationMatrix( (D3DXQUATERNION*)pQTVec1, (D3DXMATRIX*)pmGTmp );
                    vmp_QuatMatrixGeneric( pQTGVec1, *pmGTmp );
                    vmp_QuatMatrix(        pQTVec1,  *pmGTmp );

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTNorm, fPrecision ))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Matrix to Quaternion match qnorm?" << endl; 
                        cout << "    x:" << pQTNorm->x << " y:" << pQTNorm->y << " z:" << pQTNorm->z << " w:" << pQTNorm->w << "  original!!!\n";
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                    if (!vmp_IsQuatEqual( pQTGVec1, pQTVec1, fPrecision))
                      {
                        cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " Matrix to Quaternion" << endl; 
                        cout << "    x:" << pQTNorm->x << " y:" << pQTNorm->y << " z:" << pQTNorm->z << " w:" << pQTNorm->w << "  original!!!\n";
                        cout << "    x:" << pQTGVec1->x << " y:" << pQTGVec1->y << " z:" << pQTGVec1->z << " w:" << pQTGVec1->w << "  Good vs Bad!!!\n";
                        cout << "    x:" << pQTVec1->x  << " y:" << pQTVec1->y  << " z:" << pQTVec1->z  << " w:" << pQTVec1->w  << "   *** MISMATCH!!! ***\n" << endl;
                      }

                    vmp_SIMDEntry();


                        //  5th angle

                    for ( fAngV = -360.0f; fAngV <= 360.0f; fAngV += fStep )    // Yaw
                      {
                        fAngV2 = DEG2RAD( fAngV );

                            //  Quaternion Slerp (Spherical Linear Interpolation)

//                      D3DXQuaternionSlerp(  (D3DXQUATERNION*)pQTVec2,  (D3DXQUATERNION*)pQTGVec, (D3DXQUATERNION*)pQTest, fAngV2);
                        vmp_QuatSlerpGeneric( pQTGVec2, pQTGVec, pQTest, fAngV2);
                        vmp_QuatSlerp(        pQTVec2,  pQTGVec, pQTest, fAngV2);

                        if (!vmp_IsQuatEqual( pQTGVec2, pQTVec2, fPrecision))
                          {
                            cout << " " << fAngX << " " << fAngY << " " << fAngZ << " " << fAngW << " " << fAngV << " Slerp" << endl; 
                            cout << "    x:" << pQTGVec2->x << " y:" << pQTGVec2->y << " z:" << pQTGVec2->z << " w:" << pQTGVec2->w << "  Good vs Bad!!!\n";
                            cout << "    x:" << pQTVec2->x  << " y:" << pQTVec2->y  << " z:" << pQTVec2->z  << " w:" << pQTVec2->w  << "   *** MISMATCH!!! ***\n" << endl;
                          }
                      }
                  }


//              D3DXQuaternionRotationYawPitchRoll( (D3DXQUATERNION*)pQTVec, fAngY2, fAngX2, fAngZ2);
                vmp_QuatRotYawPitchRollGeneric( pQTGVec, fAngY2, fAngX2, fAngZ2 );
                vmp_QuatRotYawPitchRoll(        pQTVec,  fAngY2, fAngX2, fAngZ2 );
                
                if ( !vmp_IsQuatEqual( pQTGVec, pQTVec, fPrecision ))
                  {
                    cout << " Yaw:" << fAngY << " Pitch:" << fAngX << " Roll:" << fAngZ << endl; 
                    cout << "    x:" << pQTGVec->x << " y:" << pQTGVec->y << " z:" << pQTGVec->z << " w:" << pQTGVec->w << "  Good vs Bad!!!\n";
                    cout << "    x:" << pQTVec->x  << " y:" << pQTVec->y  << " z:" << pQTVec->z  << " w:" << pQTVec->w  << "   *** MISMATCH!!! ***\n" << endl;
                  }
              }
          }
      }
}







int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU

    cout << "\nQuaternions Test\n\n";
    cout << CpuInfoStr( szBuf, &cinfo ) << endl;

    if (!( CPUBITS_MMX & cinfo.nBits ))
      {
        cout << "ERROR! MMX Not supported!\n\n" << flush;
        return -1;
      }

    QVas3DSetup(   &cinfo );                            // QVec Add & Sub
    Matrix3DSetup( &cinfo );                            // Setup our 3D Vector stuff
    Quat3DSetup(   &cinfo );                            // Setup for Quaternions

    testQuaternions();

    cout << "Fin!\n\n" << flush;

    return 0;
}

