         TITLE Quat3DX86.asm - 80(x)86 3D Math Quaternion
   PAGE  53,160

;  Quat3D.asm - 3D (Quaternion)
;
;  This document contains file that you may use at your own risk in your own programs.
;
;  This module is designed to accomodate the book.
;
;     This module is designed for fast 3D math
;
;  Created - 3 Oct 2001 - James Leiterman
;
;  Last Modification
;
;  Tabs = 8



   .XLIST

;;;        .387
;;;   .586P   .586    .686   .686P
;;;      .K3D    .XMM    .MMX

   .686
   .XMM

   .model flat, C


        include KariType.inc
        include vmp3DX86M.inc
 
   .LIST


   .data
   align 16

;;;                 X         Y         Z         W
vSmall   DWORD 00008000h,00008000h,00008000h,00008000h

vTiny REAL4 0.0000001, 0.0000001, 0.0000001, 0.0000001
vZero REAL4 0.0,       0.0,       0.0,       0.0
vOne  REAL4 1.0,       1.0,       1.0,       1.0
vOneW REAL4 0.0,       0.0,       0.0,       1.0
vNegOne  REAL4 -1.0,     -1.0,      -1.0,      -1.0

vQQZ  DWORD 000000000h,080000000h,000000000h,080000000h  ; -w +z -y +x
vQQY  DWORD 080000000h,000000000h,000000000h,080000000h  ; -w +z +y -x
vQQX  DWORD 000000000h,000000000h,080000000h,080000000h  ; -w -z +y +x

vNegW DWORD 000000000h,000000000h,000000000h,080000000h  ;+++-
vNeg  DWORD 080000000h,080000000h,080000000h,080000000h
conjugate DWORD   080000000h,080000000h,080000000h,000000000h  ;XYZW {- - - +}
himsk32 DWORD  000000000h,000000000h,000000000h,0FFFFFFFFh
lomsk96 DWORD  0FFFFFFFFh,0FFFFFFFFh,0FFFFFFFFh,000000000h

foo   REAL4 0.0,0.0,0.0,0.0

;------------------------------------------------------------------------------------------------


Quat3DX86   SEGMENT USE32 PUBLIC 'CODE'

;-------------------------------------------------------------------------------------------------



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quaternion) Conjugate
;
;void vmp_QuatConjugate(vmpQuat * const pvD, const vmpQuat * const pvA)
;{
;  pvD->x = -pvA->x;
;  pvD->y = -pvA->y;
;  pvD->z = -pvA->z;
;  pvD->w =  pvA->w;
;
;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatConjugateAsm3DNow
vmp_QuatConjugateAsm3DNow       proc    near , vD:ptr, vA:ptr
   .K3D
;;;        femms
ASSERT_PTR4(vA)
ASSERT_PTR4(vD)

   mov   eax,vA            ; Vector A
   mov   edx,vD            ; Vector Destination

   movq  mm0,[eax+0]    ;vA.xy   {Ay Ax}
   movq  mm1,[eax+8]    ;vA.zw   {Aw Az}

   pxor  mm0,qword ptr conjugate+0
   pxor  mm1,qword ptr conjugate+8

   movq  [edx+0],mm0    ; {-Ay -Ax} 
   movq  [edx+8],mm1    ; { Aw -Az}

;;;        femms
        ret
vmp_QuatConjugateAsm3DNow endp



;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
   align 16
   public   vmp_QuatConjugateAsmSSE
vmp_QuatConjugateAsmSSE       proc    near , vD:ptr, vA:ptr
   .XMM
;;;        emms
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

   mov   eax,vA            ; Vector A
   mov   edx,vD            ; Vector D

   movaps   xmm0,[eax]        ;vA.xyzw {Aw  Az  Ay  Ax}
   xorps xmm0,oword ptr conjugate   
   movaps   [edx],xmm0        ;        {Aw -Az -Ay -Ax}  
   
;;;        emms
        ret
vmp_QuatConjugateAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;  (Quaternion) Dot Product
;
;void vmp_QuatDotProduct(float * const pfD, const vmpQuat * const pqA, const vmpQuat * const pqB)
;{ 
;  = pqA->x * pqB->x + pqA->y * pqB->y + pqA->z * pqB->z + pqA->w * pqB->w;   =AxBx+AyBy+AzBz+AwBw
;
;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatDotProductAsm3DNow
vmp_QuatDotProductAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr
   .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

   mov   ecx,vB            ; Vector B
   mov   eax,vA            ; Vector A

   movq  mm0,[ecx+8]       ;vB.zw  {Bw Bz}
   movq  mm1,[eax+8]       ;vA.zw  {Aw Az}
   movq  mm2,[ecx+0]       ;vB.xy  {By Bx}
   movq  mm3,[eax+0]       ;vA.xy  {Ay Ax}

   mov   edx,vD            ; Vector Destination

   pfmul mm0,mm1           ;*vB.z  {BwAw BzAz}
   pfmul mm2,mm3           ;       {ByAy BxAx}

   pfacc mm0,mm2           ; {ByAy+BxAx BwAw+BzAz}
   pfacc mm0,mm0           ; {#  ByAy+BxAx+BwAw+BzAz}

   movd  [edx],mm0         ; Save resulting scalar
   
;;;        femms
        ret
vmp_QuatDotProductAsm3DNow endp


;_________________________________________________________________________________________________

; *** 128bit Aligned SSE version ***

   align 16
   public   vmp_QuatDotProductAsmSSE
vmp_QuatDotProductAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr
   .XMM
;;;        emms
ASSERT_PTR(vD)
ASSERT_PTR16(vA)
ASSERT_PTR16(vB)

   mov   ecx,vB            ; Vector B
   mov   eax,vA            ; Vector A

   movaps   xmm1,[ecx]        ;vB.xyzw {Bw Bz By Bx}
   movaps   xmm0,[eax]        ;vA.xyzw {Aw Az Ay Ax}

   mov   edx,vD            ; float D
   mulps xmm0,xmm1         ; {AwBw AzBz AyBy AxBx}
   movaps   xmm1,xmm0         ; {AwBw AzBz AyBy AxBx}
   movaps   xmm2,xmm0         ; {AwBw AzBz AyBy AxBx}

   shufps   xmm1,xmm1,01001110b  ; 1 0 3 2    {AyBy AxBx AwBw AzBz}
   addps xmm1,xmm0         ; {AyBy+AwBw AxBx+AzBz AwBw+AyBy AzBz+AxBx}
   shufps   xmm0,xmm1,11111101b  ; 3 3 3 1   ; {AyBy+AwBw AxBx+AzBz AzBz+AxBx AwBw+AyBy}

   addss xmm0,xmm1         ; {# # # AzBz+AxBx+AwBw+AyBy}

   movss [edx],xmm0        ; Save Scalar

;;;        emms
        ret
vmp_QuatDotProductAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;  (Quaternion) Magnitude - (3D Distance from {0,0,0,0})
;
;
;void vmp_QuatMagnitude(float * const pfD, const vmpQuat * const pvA)
;{
;  *pfD = sqrtf( pqA->x * pqA->x + pqA->y * pqA->y + pqA->z * pqA->z + pqA->w * pqA->w );
;}
;
;_________________________________________________________________________________________________
; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatMagnitudeFastAsm3DNow
vmp_QuatMagnitudeFastAsm3DNow       proc    near , pfD:ptr, pvA:ptr
   .K3D
;;;        femms
ASSERT_PTR(pfD)
ASSERT_PTR(pvA)

   mov   eax,pvA           ; Vector A
   mov   edx,pfD           ; Vector Destination

   ;  Calculate XYZW sum of squares

   movq  mm0,[eax+0]       ;vA.xy   {Ay   Ax}
   movq  mm1,[eax+8]       ;vA.zw   {Aw   Az}

   pfmul mm0,mm0           ;       {AyAy AxAx}
   pfmul mm1,mm1           ;       {AwAw AzAz}

   pfacc mm0,mm1           ; {AwAw+AzAz AyAy+AxAx}
   pfacc mm0,mm0           ; {#  AwAw+AzAz+AyAy+AxAx}

   ;  mm0=AwAw+AzAz+AyAy+AxAx
   ;  Calculate Square Root upon Sum of squares
   ;  pfrsqrt=15bit accuracy
   pfrsqrt     mm1,mm0           ;X0=f(x)=1/sqrt(x)  {1/sqrt(r) 1/sqrt(r)}
   pfmul       mm0,mm1        ;    {r/sqrt(r) r/sqrt(r)}
                  ; r = r/sqrt(r)
   movd  [edx],mm0         ; Save resulting distance float
   
;;;        femms
        ret
vmp_QuatMagnitudeFastAsm3DNow endp

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatMagnitudeAsm3DNow
vmp_QuatMagnitudeAsm3DNow       proc    near , pfD:ptr, pvA:ptr
   .K3D
;;;        femms
ASSERT_PTR(pfD)
ASSERT_PTR(pvA)

   mov   eax,pvA           ; Vector A
   mov   edx,pfD           ; Vector Destination

   ;  Calculate XYZW sum of squares

   movq  mm0,[eax+0]       ;vA.xy   {Ay   Ax}
   movq  mm1,[eax+8]       ;vA.zw   {Aw   Az}

   pfmul mm0,mm0           ;       {AyAy AxAx}
   pfmul mm1,mm1           ;       {AwAw AzAz}

   pfacc mm0,mm1           ; {AwAw+AzAz AyAy+AxAx}
   pfacc mm0,mm0           ; {#  AwAw+AzAz+AyAy+AxAx}

   ;  mm0=AwAw+AzAz+AyAy+AxAx
   ;  Calculate Square Root upon Sum of squares
   ;  pfrsqrt=15bit accuracy
   pfrsqrt     mm1,mm0           ;X0=f(x)=1/sqrt(x)  {#####  1/sqrt(AzAz+AyAy+AxAx)}

    ; refine Newton-Raphson reciprocal square approximation
    ; Calculate 1/sqrt() accurate to 24bits
   movq        mm2,mm1     ;            {#####  1/sqrt(AzAz+AyAy+AxAx)}
   pfmul       mm1,mm1     ;X1=f(x0,x0) {#####  1/(AzAz+AyAy+AxAx)}
   pfrsqit1    mm1,mm0     ;X2=f(x,x1)     {1st step}

    ; Calculate sqrt() = 1/(1/sqrt())  24bit
   pfrcpit2    mm1,mm2     ; y=f(x2,x0)    {2nd step}
   pfmul       mm0,mm1

   movd  [edx],mm0         ; Save resulting distance float
   
;;;        femms
        ret
vmp_QuatMagnitudeAsm3DNow endp


;_________________________________________________________________________________________________

; *** 128bit Aligned SSE version ***

   align 16
   public   vmp_QuatMagnitudeAsmSSE
vmp_QuatMagnitudeAsmSSE       proc    near , pfD:ptr, pvA:ptr
   .XMM
;;;        emms
ASSERT_PTR(pfD)
ASSERT_PTR(pvA)

   mov   eax,pvA           ; Vector A
   mov   edx,pfD           ; float D

   ;  Calculate Sum of squares

   movaps   xmm0,[eax]        ;vA.xyzw {Aw Az Ay Ax}

   mulps xmm0,xmm0         ; {AwAw AzAz AyAy AxAx}

   movaps   xmm1,xmm0         ; {AwAw AzAz AyAy AxAx}
   shufps   xmm0,xmm0,01001110b ; 1 0 3 2     {AyAy AxAx AwAw AzAz}

   addps xmm0,xmm1         ; {AyAy+AwAw AxAx+AzAz AwAw+AyAy AzAz+AxAx}

   movaps   xmm1,xmm0
   shufps   xmm0,xmm1,10110001b ; 2 3 0 1     {AxAx+AzAz AyAy+AwAw AzAz+AxAx AwAw+AyAy}

   addss xmm0,xmm1         ; {  #  #  #           AzAz+AxAx+AwAw+AyAy} 

   ;  Calculate Square Root upon Sum of squares  xmm0

   sqrtss   xmm0,xmm0         ; SqRoot
   movss [edx],xmm0        ; Save 

;;;        emms
        ret
vmp_QuatMagnitudeAsmSSE endp


;_________________________________________________________________________________________________


; *** 128bit Aligned SSE version ***

   align 16
   public   vmp_QuatMagnitudeFastAsmSSE
vmp_QuatMagnitudeFastAsmSSE       proc    near , pfD:ptr, pvA:ptr
   .XMM
;;;        emms
ASSERT_PTR(pfD)
ASSERT_PTR(pvA)

   mov   eax,pvA           ; Vector A
   mov   edx,pfD           ; float D

   ;  Calculate Sum of squares

   movaps   xmm0,[eax]        ;vA.xyzw {Aw Az Ay Ax}

   mulps xmm0,xmm0         ; {AwAw AzAz AyAy AxAx}

   movaps   xmm1,xmm0         ; {AwAw AzAz AyAy AxAx}
   shufps   xmm0,xmm0,01001110b ; 1 0 3 2     {AyAy AxAx AwAw AzAz}

   addps xmm0,xmm1         ; {AyAy+AwAw AxAx+AzAz AwAw+AyAy AzAz+AxAx}

   movaps   xmm1,xmm0
   shufps   xmm0,xmm1,10110001b ; 2 3 0 1     {AxAx+AzAz AyAy+AwAw AzAz+AxAx AwAw+AyAy}

   addss xmm0,xmm1         ; {  #  #  #           AzAz+AxAx+AwAw+AyAy} 

;  Calculate Square Root upon sum of squares

   movaps   xmm2,xmm0
   movaps   xmm3,xmm0
   movaps   xmm1,xmm0

   cmpss xmm2,vTiny,5   ; >=  1's = #'s Okay
   cmpss xmm3,vTiny,1   ; <   1's = too close to zero

   ;  vD = (1 / sqrt(vA))
   rsqrtss  xmm0,xmm0

   ;  Correct for infinity due to 1/0 =  1/sqrt(0)

   andps xmm3,vOne   ; Preserve 1.0's (in positions that were infinite!)
   andps xmm0,xmm2   ; Preserve #'s that are not too small, else zeros
   orps  xmm0,xmm3   ; Blended #'s and 1.0's

   mulss xmm0,xmm1   ; {  ... Ax*1/sqrt(Ax)}

   movss [edx],xmm0        ; Save 

;;;        emms
        ret
vmp_QuatMagnitudeFastAsmSSE endp




;_________________________________________________________________________________________________
;=================================================================================================
;  (Quaternion) Normalization
;
;
;void vmp_QuatNormalize(vmpQuat * const pqD, const vmpQuat * const pqA)
;{
;  float fMag = sqrtf( pqA->x * pqA->x + pqA->y * pqA->y + pqA->z * pqA->z + pqA->w * pqA->w );
;
;  if ( fMag < 0.0001f ) { // If too close to zero? Don't want a divide by zero!
;         fMag = 1.0;
;  } else {       // Not zero so 1/n
;    fMag = 1.0f / fMag;
;  }
;
;  pqD->x = pqA->x * fMag;
;  pqD->y = pqA->y * fMag;
;  pqD->z = pqA->z * fMag;
;  pqD->w = 1.0f;
;
;  if (0.0f > pqA->w)   // get the sign!
;  {
;    pqD->w = -1.0f; // Need signed 1 or -1
;  }
;}
;
;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatNormalizeFastAsm3DNow
vmp_QuatNormalizeFastAsm3DNow       proc    near , vD:ptr, vA:ptr
   .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)

   mov   ecx,vA            ; Vector A
   mov   eax,vD            ; Vector Destination

   ;  Calculate XYZW sum of squares

   movq  mm2,[eax+0]       ;vA.xy   {Ay   Ax}
   movq  mm3,[eax+8]       ;vA.zw   {Aw   Az}
   movq  mm0,mm2
   movq  mm1,mm3

   pfmul mm2,mm2           ;       {AyAy AxAx}
   pfmul mm3,mm3           ;       {AwAw AzAz}

   pfadd mm2,mm3           ; {AwAw+AzAz AyAy+AxAx}
   pfacc mm2,mm2           ; r={#  AwAw+AzAz+AyAy+AxAx}

   ;  mm2=    r=AwAw+AzAz+AyAy+AxAx  (Sum of squares)

   ;  Calculate Square Root upon Dot Product
   ;  pfrsqrt=15bit accuracy

   ; too close zero ...???  1.0 / 10000.0 ???

   movq  mm5,mm2
   pfrsqrt  mm2,mm2           ;X0=f(x)=1/sqrt(x) {1/sqrt(r) 1/sqrt(r)}
   pfcmpgt  mm5,qword ptr vSmall              ; ( r > vSmall) ? -1 : 0

   ;  mm2=1/sqrt(AwAw+AzAz+AyAy+AxAx)

   pand  mm0,mm5           ; Not zero'ish = Keep, else zero value
   pand  mm1,mm5

   pfmul mm0,mm2           ;{Ny Nx}= {Ay/m Ax/m}
   pfmul mm1,mm2           ;{Nw Nz}= {Aw/m Az/m}

   ;  Save Resulting {w z y x} Normals
   movq  [edx+0],mm0       ; {Ny Nx}
   movq  [edx+8],mm1       ; {Nw Nz}
   
;;;        femms
        ret
vmp_QuatNormalizeFastAsm3DNow endp

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatNormalizeAsm3DNow
vmp_QuatNormalizeAsm3DNow       proc    near , vD:ptr, vA:ptr
   .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)

   mov   eax,vA            ; Vector A
   mov   edx,vD            ; Vector Destination

   ;  Calculate XYZW sum of squares

   movq  mm2,[eax+0]       ;vA.xy   {Ay   Ax}
   movq  mm3,[eax+8]       ;vA.zw   {Aw   Az}
   movq  mm0,mm2
   movq  mm1,mm3

   pfmul mm2,mm2           ;       {AyAy AxAx}
   pfmul mm3,mm3           ;       {AwAw AzAz}

   pfadd mm2,mm3           ; {AwAw+AzAz AyAy+AxAx}
   pfacc mm2,mm2           ; r={#  AwAw+AzAz+AyAy+AxAx}

   ;  mm2=    r=AwAw+AzAz+AyAy+AxAx  (Sum of squares)

   ;  Calculate Square Root upon sum of squares
   ;  pfrsqrt=15bit accuracy

   ; too close zero ...???  1.0 / 10000.0 ???

   movq  mm5,mm2
   pfrsqrt  mm4,mm2           ;X0=f(x)=1/sqrt(x) {1/sqrt(r) 1/sqrt(r)}
   pfcmpgt  mm5,qword ptr vSmall              ; ( r > vSmall) ? -1 : 0

   ;  mm2=1/sqrt(AwAw+AzAz+AyAy+AxAx)

   movq  mm3,mm4           ; 1/sqrt(x)
   pfmul mm4,mm4           ; 1/x = (1/sqrt(x))*(1/sqrt(x))

   pand  mm0,mm5           ; Not zero'ish = Keep, else zero value

   pfrsqit1 mm2,mm4
   pfrcpit2 mm2,mm3

   pand  mm1,mm5

   pfmul mm0,mm2           ;{Ny Nx}= {Ay/m Ax/m}
   pfmul mm1,mm2           ;{Nw Nz}= {Aw/m Az/m}

   ;  Save Resulting {w z y x} Normals
   movq  [edx+0],mm0       ; {Ny Nx}
   movq  [edx+8],mm1       ; {Nw Nz}
   
;;;        femms
        ret
vmp_QuatNormalizeAsm3DNow endp


;_________________________________________________________________________________________________

; *** 128bit Aligned SSE version ***

   align 16
   public   vmp_QuatNormalizeAsmSSE
vmp_QuatNormalizeAsmSSE       proc    near , pvD:ptr, pvA:ptr
   local uflow:DWORD

   .XMM
;;;        emms
ASSERT_PTR(pvD)
ASSERT_PTR(pvA)

   mov   eax,pvA           ; Vector A
   mov   edx,pvD           ; Vector D

   ;  Calculate XYZW sum of squares

   movaps   xmm6,[eax]        ;vA.xyzw {Aw Az Ay Ax}

   movaps   xmm0,xmm6         ; {Aw   Az   Ay   Ax}
   mulps xmm6,xmm6         ; {AwAw AzAz AyAy AxAx}

   movaps   xmm1,xmm6         ; {AwAw AzAz AyAy AxAx}
   shufps   xmm6,xmm6,01001110b ; 1 0 3 2     {AyAy AxAx AwAw AzAz}

   addps xmm6,xmm1         ; {AyAy+AwAw AxAx+AzAz AwAw+AyAy AzAz+AxAx}

   movaps   xmm1,xmm6         ; {AyAy+AwAw AxAx+AzAz AwAw+AyAy AzAz+AxAx}
   shufps   xmm6,xmm1,10110001b ; 2 3 0 1     {AxAx+AzAz AyAy+AwAw AzAz+AxAx AwAw+AyAy}

   addps xmm1,xmm6         ; {  r   r   r   r= AzAz+AxAx+AwAw+AyAy} 

   ;  Calculate Square Root upon Sum of squares  xmm1, xmm6=vA
   ; Sum of squares too close zero ...???  1.0 / 10000.0 ???

   movaps   xmm4,xmm0         ; {Aw   Az   Ay   Ax}
   movaps   xmm2,xmm1         ; { r    r    r    r}
   movaps   xmm3,xmm1         ; { r    r    r    r}

   andps xmm4,vNegW        ; { +-   +    +    +} Keep sign only of Aw

   cmpps xmm2,oword ptr vTiny,1     ; d = (a<b)?  -1 : 0  1's mask
   cmpps xmm3,oword ptr vTiny,5     ; d = (a>=b)? -1 : 0  #'s mask

   andps xmm2,vOneW        ;^{ 1    0    0    0} Near zero, set W to 1.0

   ;  Calculate Square Root upon Sum of squares

   sqrtps   xmm6,xmm1         ; SqRoot  {  rMw    rMz    rMy    rMx}
   divps xmm0,xmm6         ;         {Aw/Nw Az/rNz Ay/rNy Ax/rNx}

   orps  xmm2,xmm4         ;   if vA.? neg, then set to -1 else 1
   andps xmm0,xmm3         ;       NOT, mask to keep
   orps  xmm0,xmm2         ; Reblended

   movaps   [edx],xmm0        ; Save xyzw
;;;        emms
        ret
vmp_QuatNormalizeAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;  (Quaternion) Inverse
;
;
;void vmp_QuatInv(vmpQuat * const pqD, const vmpQuat * const pqA)
;{
;     // Sum of squares
;  float fLen = pqA->x * pqA->x + pqA->y * pqA->y + pqA->z * pqA->z + pqA->w * pqA->w;
;
;  if (0.0f != fLen)          // NOT absolute zero?
;    {
;     fLen = 1.0f / fLen;     // 1/r
;     pqD->x = -pqA->x * fLen;
;     pqD->y = -pqA->y * fLen;
;     pqD->z = -pqA->z * fLen;
;     pqD->w =  pqA->w * fLen;
;    }
;  else              // (INVALID {0,0,0,0} not a quaternion)
;    {
;     pqD->x = pqD->y = pqD->z = 0.0f;
;     pqD->w = 1.0f;
;    }
;
;}
;
;  NOTE: vA, vB most likely will be misaligned especially if a packed structure
;  such as VMP3DVector is used so each field must be loaded individually!
;
;_________________________________________________________________________________________________


; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

   align 16
   public   vmp_QuatInvAsm3DNow
vmp_QuatInvAsm3DNow       proc    near , vD:ptr, vA:ptr
   .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)

   mov   eax,vA            ; Vector A
   mov   edx,vD            ; Vector Destination

   ;  Calculate XYZW sum of squares

   movq  mm2,[eax+0]       ;vA.xy   {Ay   Ax}
   movq  mm3,[eax+8]       ;vA.zw   {Aw   Az}
   movq  mm0,mm2           ;vA.xy   {Ay   Ax}
   movq  mm1,mm3           ;vA.zw   {Aw   Az}

   pfmul mm2,mm2           ;       {AyAy AxAx}
   pfmul mm3,mm3           ;       {AwAw AzAz}

   pfadd mm2,mm3           ; {AwAw+AzAz AyAy+AxAx}
   pfacc mm2,mm2           ; r={#  AwAw+AzAz+AyAy+AxAx}

   ;  mm2= {r r}     r=AwAw+AzAz+AyAy+AxAx  (Sum of squares)

   ;  Calculate Square Root upon Dot Product
   ;  pfrsqrt=15bit accuracy

   ; too close zero ...???  1.0 / 10000.0 ???

   movq  mm3,mm2     ; {r r}
   pfrcp mm4,mm2           ;X0=f(x)= {1/x 1/x}

   pxor  mm0,qword ptr conjugate+0     ; {-y -x}
   pxor  mm1,qword ptr conjugate+8     ; { w -z}

   pfcmpgt  mm3,qword ptr vZero              ; ( r = 0) ? -1 : 0

   ;  mm2=1/sqrt(AwAw+AzAz+AyAy+AxAx)

   pfrcpit1 mm2,mm4

   pand  mm0,mm3           ; Not zero'ish = Keep, else zero value

   pfrcpit2 mm2,mm4

   pand  mm1,mm3

   pfmul mm0,mm2           ;{Ny Nx}= {Ay/m Ax/m}
   pfmul mm1,mm2           ;{Nw Nz}= {Aw/m Az/m}

   ;  Save Resulting {w z y x} Normals
   movq  [edx+0],mm0       ; {Ny Nx}
   movq  [edx+8],mm1       ; {Nw Nz}
   
;;;        femms
        ret
vmp_QuatInvAsm3DNow endp


;_________________________________________________________________________________________________

; *** 128bit Aligned SSE version ***

   align 16
   public   vmp_QuatInvAsmSSE
vmp_QuatInvAsmSSE       proc    near , vD:ptr, vA:ptr
   local uflow:DWORD

   .XMM
;;;        emms
ASSERT_PTR(vD)
ASSERT_PTR(vA)

   mov   ecx,vA            ; Vector A
   mov   eax,vD            ; Vector D

   ;  Calculate Dot Product

   movaps   xmm0,[ecx]        ;vA.xyz# {# Az Ay Ax}
   andps xmm0,lomsk96         ;        {0 Az Ay Ax}

   movaps   xmm6,xmm0         ; {0   Az   Ay   Ax}
   mulps xmm0,xmm0         ; {0 AzAz AyAy AxAx}

   movaps   xmm1,xmm0         ; {0 AzAz AyAy AxAx}
   movaps   xmm2,xmm0         ; {0 AzAz AyAy AxAx}

   orps  xmm1,vOneW        ; {1 AzAz AyAy AxAx}

   shufps   xmm1,xmm1,11001001b   ;3 0 2 1   ; {1 AxAz AzAz AyAy}
   shufps   xmm2,xmm2,11010010b   ;3 1 0 2   ; {0 AyAy AxAx AzAz}

   addps xmm1,xmm0         ; {1+0 AzAz+AxAx AyAy+AzAz AxAx+AyAy}
   addps xmm1,xmm2    ; {1+0+0 AzAz+AxAx+AyAy AyAy+AzAz+AxAx AxAx+AyAy+AzAz}

   ; Dot Product too close zero ...???  1.0 / 10000.0 ???

   movss uflow,xmm1
   cmp   uflow,FLOAT0001         ; 0.0001f    38d1b717h
   jl short zmag        ;just set vD=vA!!!

   ;  Calculate Square Root upon Dot Product

   sqrtps   xmm0,xmm1         ; SqRoot  {1 rMz rMy rMx}
   divps xmm6,xmm0         ; n/m     {0 rNz rNy rNx}

zmag: andps xmm6,lomsk96         ;{0 Nz Ny Nx}
   movaps   [eax],xmm6        ; Save xyz0

;;;        emms
        ret
vmp_QuatInvAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;void vmp_QuatMul(vmpQuat * const pqD, const vmpQuat * const pqA, const vmpQuat * const pqB)
;{
;  vmpQuat  quat;
;
;  quat.x = pqA->x * pqB->w + pqA->y * pqB->z - pqA->z * pqB->y + pqA->w * pqB->x;
;  quat.y = pqA->y * pqB->w - pqA->x * pqB->z + pqA->w * pqB->y + pqA->z * pqB->x;
;  quat.z = pqA->z * pqB->w + pqA->w * pqB->z + pqA->x * pqB->y - pqA->y * pqB->x;
;  pqD->w = pqA->w * pqB->w - pqA->z * pqB->z - pqA->y * pqB->y - pqA->x * pqB->x;
;  pqD->x = quat.x;
;  pqD->y = quat.y;
;  pqD->z = quat.z;
;}
;_________________________________________________________________________________________________


;  3DNow! Version - (UN)Aligned

   public   vmp_QuatMulAsm3DNow
vmp_QuatMulAsm3DNow  proc  near

   .K3D

ASSERT_PTR16( DWORD PTR [esp+8] )   ; aMx
ASSERT_PTR16( DWORD PTR [esp+4] )   ; dMx
ASSERT_PTR16( DWORD PTR [esp+12] )  ; bMx

   mov   eax,[esp+8]       ;pqA
   mov   ecx,[esp+12]      ; pqB
   mov   edx,[esp+4]       ; pqD

   movq  mm0,[eax+0]    ; {Ay Ax}
   movq  mm7,[eax+8]    ; {Aw Az}
   movq  mm1,[ecx+8]    ; {Bw Bz}
   movq  mm6,[ecx+0]    ; {By Bx}
   movq  mm5,mm1        ; {Bw Bz}
   movq  mm4,mm6        ; {By Bx}

   punpckhdq   mm1,mm1     ; {Bw Bw}
   punpckldq   mm5,mm5     ; {Bz Bz}
   punpckhdq   mm6,mm6     ; {By By}
   punpckldq   mm4,mm4     ; {Bx Bx}

   movq  mm2,mm0        ; {Ay Ax}
   pfmul mm0,mm1        ; {AyBw AxBw}
   pfmul mm1,mm7        ; {AwBw AzBw}

   movq  mm3,mm2        ; {Ay Ax}
   pfmul mm3,mm6        ; {AyBy AxBy}
   pfmul mm6,mm7        ; {AwBy AzBy}
   pxor  mm3,qword ptr vQQY+8 ; {-AyBy +AxBy}
   pxor  mm6,qword ptr vQQY+0 ; {+AwBy -AzBy}

   pfadd mm0,mm6        ; {AyBw+AwBy AxBw-AzBy}
   pfadd mm1,mm3        ; {AwBw-AyBy AzBw+AxBy}
                                 
      ; 3DNow+ equiv: pswapd
   movq  [esp-020h],mm2 ; {Ay Ax}
   movq  [esp-018h],mm2
   movq  [esp-030h],mm7 ; {Aw Az}
   movq  [esp-028h],mm7
   movq  mm2,[esp-01ch] ; {Ax Ay}
   movq  mm7,[esp-02ch] ; {Az Aw}

      ; mm2,mm7,mm0,mm1  mm5,mm4  

   movq  mm3,mm2        ; {Ax Ay}
   pfmul mm3,mm5        ; {AxBz AyBz}
   pfmul mm5,mm7        ; {AzBz AwBz}
   pxor  mm3,qword ptr vQQZ+0 ; {-AxBz +AyBz}
   pxor  mm5,qword ptr vQQZ+8 ; {-AzBz +AwBz}

   pfadd mm0,mm3        ; {AyBw+AwBy-AxBz  AxBw-AzBy+AyBz}
   pfadd mm1,mm5        ; {AwBw-AyBy-AzBz  AzBw+AxBy+AwBz}

      ; mm2, mm7, mm0,mm1 mm4

   movq  mm3,mm2        ; {Ax Ay}
   pfmul mm2,mm4        ; {AxBx AyBx}
   pfmul mm4,mm7        ; {AzBx AwBx}

   pfadd mm0,mm4        ; {AyBw+AwBy-AxBz+AzBx  AxBw-AzBy+AyBz+AwBx}
   pfsub mm1,mm2        ; {AwBw-AyBy-AzBz-AxBx  AzBw+AxBy+AwBz-AyBx}

   movq  [edx+0],mm0
   movq  [edx+8],mm1

   ret
vmp_QuatMulAsm3DNow  endp

;_________________________________________________________________________________________________


;  SSE Version - (Aligned)

   public   vmp_QuatMulAsmSSE
vmp_QuatMulAsmSSE proc  near
   .XMM

ASSERT_PTR16( DWORD PTR [esp+8] )   ; pqA
ASSERT_PTR16( DWORD PTR [esp+12] )  ; pqB
ASSERT_PTR16( DWORD PTR [esp+4] )   ; pqD

   mov   eax,[esp+8]    ; pqA
   mov   ecx,[esp+12]      ; pqB
   mov   edx,[esp+4]    ; pqD

   movaps   xmm0,[eax]     ; {Aw Az Ay Ax}
   movaps   xmm4,[ecx]     ; {Bw Bz By Bx}
   
   movaps   xmm1,xmm0
   movaps   xmm2,xmm0
   movaps   xmm3,xmm0

   shufps   xmm1,xmm1,10110001b  ; {Az Aw Ax Ay}
   shufps   xmm2,xmm2,01001110b  ; {Ay Ax Aw Az}
   shufps   xmm3,xmm3,00011011b  ; {Ax Ay Az Aw}

   movaps   xmm5,xmm4      ; {Bw Bz By Bx}
   movaps   xmm6,xmm4      ; {Bw Bz By Bx}
   movaps   xmm7,xmm4      ; {Bw Bz By Bx}
   shufps   xmm4,xmm4,00000000b  ; {Bx Bx Bx Bx}

   shufps   xmm5,xmm5,11111111b  ; {Bw Bw Bw Bw}
   shufps   xmm6,xmm6,10101010b  ; {Bz Bz Bz Bz}
   shufps   xmm7,xmm7,01010101b  ; {By By By By}

   mulps xmm0,xmm5      ; {AwBw AzBw AyBw AxBw}
   mulps xmm1,xmm6      ; {AzBz AwBz AxBz AyBz} - + - +
   mulps xmm2,xmm7      ; {AyBy AxBy AwBy AzBy} - + + -
   mulps xmm3,xmm4      ; {AxBx AyBx AzBx AwBx} - - + +

   xorps xmm1,oword ptr vQQZ  ; {-AzBz +AwBz -AxBz +AyBz}
   xorps xmm2,oword ptr vQQY  ; {-AyBy +AxBy +AwBy -AzBy}
   xorps xmm3,oword ptr vQQX  ; {-AxBx -AyBx +AzBx +AwBx}

   addps xmm0,xmm1
   addps xmm2,xmm3
   addps xmm0,xmm2

   movaps   [edx],xmm0     ; Save quat
   ret
vmp_QuatMulAsmSSE endp




;_________________________________________________________________________________________________




Quat3DX86   ends
   end

