  
//
//  vmp3D.h - 3D Math Core Header
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to accomodate the book.
//
//      This module is designed for fast 3D math
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include    "KariType.h"                                // Kari Games Master Definitions

#ifndef __vmpCore_H
#define __vmpCore_H

#ifdef __cplusplus
extern "C" {
#endif



#define PI          (3.141592f)
#define DEG2RAD(x)  (((x) * PI) / 180.0F)
#define RAD2DEG(x)  (((x) * 180.0f) / PI)


#define DOUBLE_PRECISION    0.0000000001
#define SINGLE_PRECISION    0.0000001f
#define FAST_PRECISION      0.001f      // Division=0.001f, Sqrt=0.0001f



//
// 3D Coordinate (Single-Precision Floating-Point)
//

typedef struct vmp3DVectorType
{
    float x;
    float y;
    float z;
} vmp3DVector;



//
// 3D Coordinate Quad Vector (Single-Precision Floating-Point)
//

typedef struct vmp3DQVectorType
{
    float x;
    float y;
    float z;
    float w;
} vmp3DQVector;




//
// 3D Coordinate (Single-Precision Floating-Point)
//

typedef struct vmp3DDVectorType
{
    double x;
    double y;
    double z;
} vmp3DDVector;



//
// 3D Coordinate Quad Vector (Double-Precision Floating-Point)
//

typedef struct vmp3DQDVectorType
{
    double x;
    double y;
    double z;
    double w;
} vmp3DQDVector;




//
// Quaternion Vector (Single-Precision Floating-Point)
//

typedef struct vmpQuatType
{
    float x;
    float y;
    float z;
    float w;
} vmpQuat;



//
// Quaternion Vector (Double-Precision Floating-Point)
//

typedef struct vmpDQuatType
{
    double x;
    double y;
    double z;
    double w;
} vmpDQuat;



//
//  Texture Coordinate
//

typedef struct vmp3DVertexType
{
    float x;
    float y;
    float z;

    float nx;                                           // Normals
    float ny;
    float nz;

    float tu;                                           // Texture U-coord.
    float tv;                                           // Texture V-coord.
} vmp3DVertex;



//
//  3D Matrix
//

typedef struct vmp3DMatrix1Type
{
            float _11, _12, _13, _14;
            float _21, _22, _23, _24;
            float _31, _32, _33, _34;
            float _41, _42, _43, _44; 
} vmp3DMatrix1;         

#define MTX1(ptr)  ((vmp3DMatrix1 *)&(ptr))


typedef struct vmp3DMatrix0Type
{
            float _00, _01, _02, _03;
            float _10, _11, _12, _13;
            float _20, _21, _22, _23;
            float _30, _31, _32, _33; 
} vmp3DMatrix0;

#define MTX0(ptr)  ((vmp3DMatrix1 *)&(ptr))

                    

#ifndef CC_VMP_PS2
typedef float  vmp3DMatrix[4][4];
#else
//typedef int       IVECTOR[4] __attribute__((aligned (16)));
//typedef float     FVECTOR[4] __attribute__((aligned (16)));
typedef float  vmp3DMatrix[4][4] __attribute__((aligned (16)));
#endif





// I would love to use a union but not very portable!




bool vmp_IsFEqual(float fA, float fB, float fPrecision);
bool vmp_IsDEqual(double fA, double fB, double fPrecison);


#ifdef VMP_ADVANCED                         // Advanced? (Pre-processor Definition)

    //
    //  *** POST - Chapter 6 ***
    //

typedef bool (*vmp_IsVEqualProc)( const vmp3DVector * const pvA, const vmp3DVector * const pvB, float fPrecision );
extern vmp_IsVEqualProc vmp_IsVEqual;


typedef bool (*vmp_IsQVEqualProc)( const vmp3DQVector * const pvA, const vmp3DQVector * const pvB, float fPrecision );
extern vmp_IsQVEqualProc vmp_IsQVEqual;

    //  Quat is nearly identical to QVec just different type definition.

typedef bool (*vmp_IsQuatEqualProc)( const vmpQuat * const pqA, const vmpQuat * const pqB, float fPrecision );
extern vmp_IsQuatEqualProc vmp_IsQuatEqual;


typedef bool (*vmp_IsMatrixEqualProc)(  vmp3DMatrix aMx, vmp3DMatrix bMx, float fPrecision);
extern vmp_IsMatrixEqualProc vmp_IsMatrixEqual;


#else
    //
    //  *** PRE - Chapter 6 ***
    //

bool vmp_IsVEqualGeneric( const vmp3DVector * const pvA, const vmp3DVector * const pvB, float fPrecision );
bool vmp_IsQVEqualGeneric( const vmp3DQVector * const pvA, const vmp3DQVector * const pvB, float fPrecision );

bool vmp_IsMatrixEqualGeneric( vmp3DMatrix aMx, vmp3DMatrix bMx, float fPrecision);

bool vmp_IsQuatEqualGeneric(const vmpQuat * const pqA, const vmpQuat * const pqB, float fPrecision);


#define vmp_IsVEqual        vmp_IsVEqualGeneric
#define vmp_IsQVEqual       vmp_IsQVEqualGeneric
#define vmp_IsMatrixEqual   vmp_IsMatrixEqualGeneric

#define vmp_IsQuatEqual     vmp_IsQuatEqualGeneric


#endif




#ifdef __cplusplus
}
#endif

#endif  //__vmp3D_H


